#ifndef gb_t
#define gb_t
#include "funcs.h"
#include "mathops.h"
#include "poly.h"
#include "contain.t"
#include "ideal.t"

/////////////////
// DEFINITIONS //
/////////////////

/**************************************/
/* set Grbner basis algorithm to use */
/**************************************/

enum GB_Type {
  Buchberger     ,
  BuchbergerHomo ,
  BuchbergerGauss,
  BuchbergerMulti,
  F4
};

GB_Type GBType = Buchberger;

void   SetGBAlg(const string&);
string GetOpt();
string GetOptString();


/****************************/
/* set optimizations to use */
/****************************/

bool opt_NSS = true , // normal selection strategy
     opt_GM  = true , // Gebauer Mller installation
     opt_KR  = false; // Kreuzer Robbiano installation

string GetOpt(             );
void   SetOpt(const string&);


/*****************/
/* class pairset */
/*****************/

template<class numeric> class pairset {
  public:

  unsigned long val1, val2;
  monomial      g, l, tij, tji;
  poly<numeric> SVector;

  // constructors:
  pairset(                      ) {};
  pairset(const unsigned long &i) {};
  pairset(const unsigned long &i ,
          const unsigned long &j ,
                poly<numeric> *gi,
                poly<numeric> *gj):
    val1(i) ,
    val2(j) {
    monomial ti = gi->LT(),
             tj = gj->LT();
    g   = gcd(ti, tj);
    l   = ti * tj / g;
    tij = l / ti;
    tji = l / tj;
    SVector  = tij * (*gi / gi->LC());
    SVector -= tji * (*gj / gj->LC());
  };

  // operators:
  operator=(const pairset &p) {
    val1    = p.val1   ;
    val2    = p.val2   ;
    g       = p.g      ;
    l       = p.l      ;
    tij     = p.tij    ;
    tji     = p.tji    ;
    SVector = p.SVector;
  }
};

template<class numeric> class homopairset {
  public:

  unsigned long     val1, val2;
  monomial          g, l, tij, tji;
  homopoly<numeric> SVector;

  // constructors:
  homopairset(                           ) {};
  homopairset(const unsigned long     &i ) {};
  homopairset(const unsigned long &i ,
              const unsigned long &j ,
              homopoly<numeric>   *gi,
              homopoly<numeric>   *gj):
    val1(i) ,
    val2(j) {
    monomial ti = gi->LT(),
             tj = gj->LT();
    g   = gcd(ti, tj);
    l   = ti * tj / g;
    tij = l / ti;
    tji = l / tj;
    SVector  = tij * (*gi / gi->LC());
    SVector -= tji * (*gj / gj->LC());
  };

  // operators:
  operator=(const homopairset &p) {
    val1    = p.val1   ;
    val2    = p.val2   ;
    g       = p.g      ;
    l       = p.l      ;
    tij     = p.tij    ;
    tji     = p.tji    ;
    SVector = p.SVector;
  }
  operator==(const homopairset &p) {
    return val1 == p.val1 && val2 == p.val2;
  }

  // degree:
  deg() {
    return l.deg();
  }
};

template<class numeric> ostream& operator<<(ostream &out, const pairset<numeric> &pair) {
  out << '(' << pair.val1 << ", " << pair.val2 << ')';
  return out;
}

template<class numeric> ostream& operator<<(ostream &out, homopairset<numeric> pair) {
  out << '(' << pair.val1 << ", " << pair.val2 << ")<" << pair.deg() << '>';
  return out;
}


/***************************/
/* Grbner basis functions */
/***************************/

template<class numeric> list<    poly<numeric> > GB               (const ideal    <numeric>&, const GB_Type &gt = GBType);
template<class numeric> list<    poly<numeric> > GB               (const homoideal<numeric>&, const GB_Type &gt = GBType);
template<class numeric> list<    poly<numeric> > GBBuchberger     (const ideal    <numeric>&                            );
template<class numeric> list<homopoly<numeric> > GBBuchbergerGauss(const homoideal<numeric>&                            );
template<class numeric> list<homopoly<numeric> > GBBuchbergerHomo (const homoideal<numeric>&                            );
template<class numeric> list<homopoly<numeric> > GBBuchbergerMulti(const homoideal<numeric>&                            );
template<class numeric> list<homopoly<numeric> > F4Homo           (const homoideal<numeric>&                            );

/*******************************************/
/* Grbner basis functions with statistics */
/*******************************************/

template<class numeric> list<    poly<numeric> > StatGBBuchberger     (const    ideal <numeric>&);
template<class numeric> list<homopoly<numeric> > StatGBBuchbergerGauss(const homoideal<numeric>&);
template<class numeric> list<homopoly<numeric> > StatGBBuchbergerHomo (const homoideal<numeric>&);
template<class numeric> list<homopoly<numeric> > StatGBBuchbergerMulti(const homoideal<numeric>&);
template<class numeric> list<homopoly<numeric> > StatF4Homo           (const homoideal<numeric>&);

/*******************/
/* other functions */
/*******************/

template<class numeric   > list<homopoly<numeric> > FR           (list<homopoly   <numeric> > , list<homopoly   <numeric> >&, list<list<homopoly<numeric> > >&, list<list<homopoly<numeric> > >&                                );
template<class numeric   > list<homopoly<numeric> > FR           (list<homopoly   <numeric> > , list<homopoly   <numeric> >&, list<list<homopoly<numeric> > >&, list<list<homopoly<numeric> > >&, unsigned long&, unsigned long&);
template<class numeric   > list<homopoly<numeric> > interreduce  (list<homopoly   <numeric> >                                                                                                                                   );
template<class numeric   > list<homopoly<numeric> > interreduce  (list<homopoly   <numeric> > , unsigned long&                                                                                                                  );
template<class polynomial> bool                     isGB         (list<polynomial>&                                                                                                                                             );
template<class numeric   > sparsematrix<numeric>    MatrixUpdate (const sparsematrix<numeric>&, const unsigned long        &, list<homopoly<numeric> >&                                                                         );
template<class T         > unsigned long            MinDeg       (list<T                    >&                                                                                                                                  );
template<class numeric   > void                     MinPairs     (list<homopairset<numeric> >&, list<homopairset<numeric> >&, list<homopairset<numeric> >&    , list<homopoly<numeric> >&                                       );
template<class numeric   > void                     MR           (list<homopoly   <numeric> >&, list<homopoly   <numeric> >&, const unsigned long&                                                                              );
template<class numeric   > void                     MR           (list<homopoly   <numeric> >&, list<homopoly   <numeric> >&, const unsigned long&            , unsigned long&                                                  );
template<class numeric   > void                     pair_optimize(list<    pairset<numeric> >&, list<    poly   <numeric> >&                                                                                                    );
template<class numeric   > void                     pair_optimize(list<homopairset<numeric> >&, list<homopoly   <numeric> >&                                                                                                    );
template<class numeric   > void                     pair_update  (list<homopairset<numeric> >&, list<homopoly   <numeric> >&                                                                                                    );
template<class numeric   > void                     pair_update  (list<    pairset<numeric> >&, list<    poly   <numeric> >&                                                                                                    );
template<class pair      > void                     PairSort     (list<pair                 >&                                                                                                                                  );
template<class polynomial> void                     PolySort     (list<polynomial           >&, const unsigned long &start                                                                                                      );
template<class numeric   > homopoly<numeric>        Simp         (monomial                    , homopoly<numeric>           , list<list<homopoly<numeric> > >&, list<list<homopoly<numeric> > >&                                );
template<class numeric   > homopoly<numeric>        Simp         (monomial                    , homopoly<numeric>           , list<list<homopoly<numeric> > >&, list<list<homopoly<numeric> > >&, unsigned long&                );
template<class numeric   > void                     Update       (list<homopairset<numeric> >&, list<homopairset<numeric> >&, list<homopoly<numeric> >&                                                                         );


////////////////////
// IMPLEMENTATION //
////////////////////

void SetGBAlg(const string &type) {
  if (type == "Buchberger")
    GBType = Buchberger;
  else if (type == "BuchbergerHomo")
    GBType = BuchbergerHomo;
  else if (type == "BuchbergerGauss")
    GBType = BuchbergerGauss;
  else if (type == "BuchbergerMulti")
    GBType = BuchbergerMulti;
  else if (type == "F4")
    GBType = F4;
}

void SetOpt(const string &opt) {
  if (opt == "off") {
    opt_NSS = false;
    opt_GM  = false;
    opt_KR  = false;
  }
  else {
    opt_NSS = isin(opt, 'n') && (GBType != BuchbergerMulti) && (GBType != F4);
    opt_KR  = isin(opt, 'p') && (GBType != Buchberger);
    opt_GM  = isin(opt, 'g') && !opt_KR;
  }
}

string GetOpt() {
  string opt;
  if (opt_NSS)
    opt += 'N';
  if (opt_KR)
    opt += 'P';
  else if (opt_GM)
    opt += 'G';
  if (opt.length() == 0)
    return "";
  return opt;
}

string GetOptString() {
  string opt;
  if (opt_NSS)
    opt += ", Normal Selection Strategy";
  if (opt_KR)
    opt += ", Pair Minimalization";
  else if (opt_GM)
    opt += ", Gebauer Moeller";
  if (opt.length() == 0)
    return "none";
  return opt.substr(2);
}

template<class numeric> list<poly<numeric> > GB(const ideal<numeric> &I, const GB_Type &gt) {
  unsigned long i;
  list<homopoly<numeric> > HL;
  list<poly    <numeric> > L;
  switch(gt) {
    case Buchberger     : if (Show_Stats)
                            return StatGBBuchberger(I);
                          else
                            return GBBuchberger(I);
    case BuchbergerHomo : Homo_Error = !I.ishomo();
                          if (Homo_Error)
                            return list<poly<numeric> >();
                          if (Show_Stats)
                            HL = StatGBBuchbergerHomo(homoideal<numeric>(I));
                          else
                            HL = GBBuchbergerHomo(homoideal<numeric>(I));
                          L = list<poly<numeric> >(HL.length(), poly<numeric>());
                          for (i = 1; i <= L.length(); i++)
                            L[i] = HL[i];
                          return L;
    case BuchbergerGauss: Homo_Error = !I.ishomo();
                          if (Homo_Error)
                            return list<poly<numeric> >();
                          if (Show_Stats)
                            HL = StatGBBuchbergerGauss(homoideal<numeric>(I));
                          else
                            HL = GBBuchbergerGauss(homoideal<numeric>(I));
                          L = list<poly<numeric> >(HL.length(), poly<numeric>());
                          for (i = 1; i <= L.length(); i++)
                            L[i] = HL[i];
                          return L;
    case BuchbergerMulti: Homo_Error = !I.ishomo();
                          if (Homo_Error)
                            return list<poly<numeric> >();
                          if (Show_Stats)
                            HL = StatGBBuchbergerMulti(homoideal<numeric>(I));
                          else
                            HL = GBBuchbergerMulti(homoideal<numeric>(I));
                          L = list<poly<numeric> >(HL.length(), poly<numeric>());
                          for (i = 1; i <= L.length(); i++)
                            L[i] = HL[i];
                          return L;
    case F4             : Homo_Error = !I.ishomo();
                          if (Homo_Error)
                            return list<poly<numeric> >();
                          if (Show_Stats)
                            HL = StatF4Homo(homoideal<numeric>(I));
                          else
                            HL = F4Homo(homoideal<numeric>(I));
                          L = list<poly<numeric> >(HL.length(), poly<numeric>());
                          for (i = 1; i <= L.length(); i++)
                            L[i] = HL[i];
                          return L;
  }
}

template<class numeric> list<poly<numeric> > GB(const homoideal<numeric> &I, const GB_Type &gt) {
  unsigned long i;
  switch(gt) {
    case Buchberger     : if (Show_Stats)
                            return StatGBBuchberger(ideal<numeric>(I));
                          else
                            return GBBuchberger(ideal<numeric>(I));
    case BuchbergerHomo : Homo_Error = false;
                          list<homopoly<numeric> > HL;
                          if (Show_Stats)
                            HL = StatGBBuchbergerHomo(I);
                          else
                            HL = GBBuchbergerHomo(I);
                          list<poly<numeric> > L(HL.length());
                          for (i = 1; i <= L.length(); i++)
                            L[i] = HL[i];
                          return L;
    case BuchbergerGauss: Homo_Error = false;
                          list<homopoly<numeric> > HL;
                          if (Show_Stats)
                            HL = StatGBBuchbergerGauss(I);
                          else
                            HL = GBBuchbergerGauss(I);
                          list<poly<numeric> > L(HL.length());
                          for (i = 1; i <= L.length(); i++)
                            L[i] = HL[i];
                          return L;
    case BuchbergerMulti: Homo_Error = false;
                          list<homopoly<numeric> > HL;
                          if (Show_Stats)
                            HL = StatGBBuchbergerMulti(I);
                          else
                            HL = GBBuchbergerMulti(I);
                          list<poly<numeric> > L(HL.length());
                          for (i = 1; i <= L.length(); i++)
                            L[i] = HL[i];
                          return L;
    case F4             : Homo_Error = false;
                          list<homopoly<numeric> > HL;
                          if (Show_Stats)
                            HL = StatF4Homo(I);
                          else
                            HL = F4Homo(I);
                          list<poly<numeric> > L(HL.length());
                          for (i = 1; i <= L.length(); i++)
                            L[i] = HL[i];
                          return L;
  }
}

/************************/
/* Buchberger algorithm */
/************************/

template<class numeric> list<poly<numeric> > GBBuchberger(const ideal<numeric> &I) {
  list<pairset<numeric> > B;
  list<poly   <numeric> > G = I.gens;
  monomial                g, one;
  poly        <numeric>   nr, S, zeropoly;
  unsigned long i, j, k, t = G.length();

  for (i = 1; i < t; i++)
    for (j = i + 1; j <= t; j++)
      B.append(pairset<numeric>(i, j, &G[i], &G[j]));
  if (opt_NSS)
    PairSort(B);

  while (B.length() > 0) {
    g = B[1].g;
    S = B[1].SVector;
    B.remove(1);
    if (g != one && S != zeropoly) {
      nr = S.NR(G);
      if (nr != zeropoly) {
        t++;
        G.append(nr);
        if (opt_GM)
          pair_optimize(B, G);
        else
          pair_update(B, G);
        if (opt_NSS)
          PairSort(B);
      }
    }
  }

  return G;
}

/****************************************/
/* Buchberger algorithm with statistics */
/****************************************/

template<class numeric> list<poly<numeric> > StatGBBuchberger(const ideal<numeric> &I) {
  list<pairset<numeric> > B;
  list<poly   <numeric> > G = I.gens;
  monomial                g, one;
  poly<numeric>           nr, S, zeropoly;
  string        comp_time;
  unsigned long i, j, k, t = G.length(),
                nr_count = 0, nr_steps   = 0, nr_zero   = 0,
                pair_all = 0, pair_count = 0, pair_used = 0;

  timer('s');
  for (i = 1; i < t; i++)
    for (j = i + 1; j <= t; j++)
      B.append(pairset<numeric>(i, j, &G[i], &G[j]));
  if (opt_NSS)
    PairSort(B);
  pair_count = B.length();

  while (B.length() > 0) {
    pair_used++;
    g = B[1].g;
    S = B[1].SVector;
    B.remove(1);
    if (g != one && S != zeropoly) {
      nr_count++;
      nr = S.StatNR(G, nr_steps);
      if (nr == zeropoly)
        nr_zero++;
      else{
        t++;
        G.append(nr);
        pair_all += t - 1;
        i = B.length();
        if (opt_GM)
          pair_optimize(B, G);
        else
          pair_update(B, G);
        pair_count += B.length() - i;
        if (opt_NSS)
          PairSort(B);
      }
    }
  }

  comp_time = FormatTime(timer('e'));
  cout << endl
       << replicate('', 79)                                                        << endl
       << "algorithm:              Buchberger"                                      << endl
       << "optimizations:          " << GetOptString()                              << endl
       << replicate('', 79)                                                        << endl
       << "number of all    pairs: " << pair_all                                    << endl
       << "number of listed pairs: " << pair_count                                  << endl
       << "number of used   pairs: " << pair_used                                   << endl
       << "reduction calls:        " << nr_count << " (" << nr_zero << " vanished)" << endl
       << "reduction steps:        " << nr_steps                                    << endl
       << "original generators:    " << I.gens.length()                             << endl
       << "GB       generators:    " << G.length()                                  << endl
       << "new      generators:    " << G.length() - I.gens.length()                << endl
       << "computation time:       " << comp_time                                   << endl
       << "GB verification:        " << (isGB(G) ? "ok" : "failure")                << endl
       << replicate('', 79)                                                        << endl
       << endl;

  return G;
}

/***********************************************/
/* Buchberger algorithm for homogeneous ideals */
/***********************************************/

template<class numeric> list<homopoly<numeric> > GBBuchbergerHomo(const homoideal<numeric> &I) {
  homopoly        <numeric>   nr, S, v, zeropoly;
  list<homopairset<numeric> > A, B, Bast, Bastd, Bd;
  list<homopoly   <numeric> > G, V = I.gens, Vd;
  monomial                    g, one;
  bool          addpair;
  unsigned long d, i, s = 0;

  if (V.length() == 0)
    return G;

  while (B.length() > 0 || V.length() > 0) {
    if (B.length() == 0)
      d = MinDeg(V);
    else if (V.length() == 0)
      d = MinDeg(B);
    else
      d = min(MinDeg(B), MinDeg(V));
    for (i = 1; i <= B.length();)
      if (B[i].deg() == d) {
        Bd.append(B[i]);
        B.remove(i);
      }
      else
        i++;
    if (opt_NSS)
      PairSort(Bd);
    for (i = 1; i <= V.length();)
      if (V[i].deg() == d) {
        Vd.append(V[i]);
        V.remove(i);
      }
      else
        i++;
    if (opt_KR) {
      for (i = 1; i <= Bast.length();)
        if (Bast[i].deg() == d) {
          Bastd.append(Bast[i]);
          Bast.remove(i);
        }
        else
          i++;
      MinPairs(A, Bd, Bastd, G);
    }
    while (Bd.length() > 0 || Vd.length() > 0) {
      addpair = true;
      if (Bd.length() > 0) {
        g = Bd[1].g;
        S = Bd[1].SVector;
        if (opt_KR)
          A.append(Bd[1]);
        Bd.remove(1);
        if (g == one || S == zeropoly)
          addpair = false;
        else {
          nr = S.NR(G);
          if (nr == zeropoly)
            addpair = false;
        }
      }
      else {
        v = Vd[1];
        Vd.remove(1);
        nr = v.NR(G);
        if (nr == zeropoly)
          addpair = false;
      }
      if (addpair) {
        s++;
        G.append(nr);
        if (opt_KR)
          Update(B, Bast, G);
        else if (opt_GM)
          pair_optimize(B, G);
        else
          pair_update(B, G);
      }
    }
  }

  return G;
}

/***************************************************************/
/* Buchberger algorithm for homogeneous ideals with statistics */
/***************************************************************/

template<class numeric> list<homopoly<numeric> > StatGBBuchbergerHomo(const homoideal<numeric> &I) {
  monomial                    g, one;
  homopoly        <numeric>   nr, S, v, zeropoly;
  list<homopairset<numeric> > A, B, Bast, Bastd, Bd;
  list<homopoly   <numeric> > G, V = I.gens, Vd;
  bool          addpair;
  string        comp_time;
  unsigned long d, dmin, i, s = 0,
                nr_count = 0, nr_steps   = 0, nr_zero   = 0,
                pair_all = 0, pair_count = 0, pair_used = 0;

  if (V.length() == 0)
    return G;

  timer('s');
  dmin = V[1].deg();
  while (B.length() > 0 || V.length() > 0) {
    if (B.length() == 0)
      d = MinDeg(V);
    else if (V.length() == 0)
      d = MinDeg(B);
    else
      d = min(MinDeg(B), MinDeg(V));
    dmin = min(dmin, d);
if (dooutput) {
cerr << endl;
cerr << "d = " << d << endl;
cerr << "G = " << G << endl;
cerr << "B = " << B << endl;
cerr << "V = " << V << endl;
}
    for (i = 1; i <= B.length();)
      if (B[i].deg() == d) {
        Bd.append(B[i]);
        B.remove(i);
      }
      else
        i++;
if (dooutput)
cerr << "B" << d << " = " << Bd << endl;
    if (opt_NSS)
      PairSort(Bd);
    for (i = 1; i <= V.length();)
      if (V[i].deg() == d) {
        Vd.append(V[i]);
        V.remove(i);
      }
      else
        i++;
if (dooutput)
cerr << "V" << d << " = " << Vd << endl;
    if (opt_KR) {
      for (i = 1; i <= Bast.length();)
        if (Bast[i].deg() == d) {
          Bastd.append(Bast[i]);
          Bast.remove(i);
        }
        else
          i++;
      MinPairs(A, Bd, Bastd, G);
    }
    while (Bd.length() > 0 || Vd.length() > 0) {
      addpair = true;
      if (Bd.length() > 0) {
if (dooutput)
cerr << "pair -" << Bd.length() << endl;
        pair_used++;
        g = Bd[1].g;
        S = Bd[1].SVector;
        if (opt_KR)
          A.append(Bd[1]);
        Bd.remove(1);
        if (g == one || S == zeropoly)
          addpair = false;
        else {
          nr_count++;
          nr = S.StatNR(G, nr_steps);
          if (nr == zeropoly) {
            nr_zero++;
            addpair = false;
          }
        }
      }
      else {
        v = Vd[1];
        Vd.remove(1);
        nr_count++;
        nr = v.StatNR(G, nr_steps);
        if (nr == zeropoly)
          addpair = false;
      }
      if (addpair) {
        s++;
        G.append(nr);
        pair_all += s - 1;
        i         = B.length();
        if (opt_KR)
          Update(B, Bast, G);
        else if (opt_GM)
          pair_optimize(B, G);
        else
          pair_update(B, G);
        pair_count += B.length() - i;
      }
    }
  }

  comp_time = FormatTime(timer('e'));
  cout << endl
       << replicate('', 79)                                                                                    << endl
       << "algorithm:              BuchbergerHomo"                                                              << endl
       << "optimizations:          " << GetOptString()                                                          << endl
       << replicate('', 79)                                                                                    << endl
       << "computed degrees: min:  " << dmin                                                                    << endl
       << "                  max:  " << d                                                                       << endl
       << "number of all    pairs: " << pair_all                                                                << endl
       << "number of listed pairs: " << pair_count                                                              << endl
       << "number of used   pairs: " << pair_used                                                               << endl
       << "reduction calls:        " << nr_count << " (" << nr_zero << " vanished)"                             << endl
       << "reduction steps:        " << nr_steps                                                                << endl
       << "original generators:    " << I.gens.length()                                                         << endl
       << "GB       generators:    " << G.length()                                                              << endl
       << "new      generators:    " << (set<homopoly<numeric> >(G) - set<homopoly<numeric> >(I.gens)).length() << endl
       << "computation time:       " << comp_time                                                               << endl
//       << "GB verification:        " << (isGB(G) ? "ok" : "failure")                                            << endl
       << replicate('', 79)                                                                                    << endl
       << endl;

  return G;
}

/***********************************************************/
/* Buchberger algorithm for homogeneous ideals using Gauss */
/***********************************************************/

template<class numeric> list<homopoly<numeric> > GBBuchbergerGauss(const homoideal<numeric> &I) {
  homopoly        <numeric>   nr, S, v, zeropoly;
  list<homopairset<numeric> > A, B, Bast, Bastd, Bd;
  list<homopoly   <numeric> > G, V = I.gens, Vd;
  monomial                    g, one;
  sparsematrix    <numeric>   M;
  unsigned long d, dmin, i, s = 0, start;

  if (V.length() == 0)
    return G;

  d    = V[1].deg();
  dmin = d;
  for (i = 2; i <= V.length(); i++) {
    d    = V[i].deg();
    dmin = min(dmin, d);
  }
  d = dmin;

  while (B.length() > 0 || V.length() > 0) {
    for (i = 1; i <= B.length();)
      if (B[i].deg() == d) {
        Bd.append(B[i]);
        B.remove(i);
      }
      else
        i++;
    if (opt_NSS)
      PairSort(Bd);
    for (i = 1; i <= V.length();)
      if (V[i].deg() == d) {
        Vd.append(V[i]);
        V.remove(i);
      }
      else
        i++;
    if (opt_KR) {
      for (i = 1; i <= Bast.length();)
        if (Bast[i].deg() == d) {
          Bastd.append(Bast[i]);
          Bast.remove(i);
        }
        else
          i++;
      MinPairs(A, Bd, Bastd, G);
    }
    M  = MatrixUpdate(M, d, G);

    if (Bd.length() > 0 || Vd.length() > 0) {
      start = M.getheight() + 1;
      while (Bd.length() > 0) {
        g = Bd[1].g;
        S = Bd[1].SVector;
        if (opt_KR)
          A.append(Bd[1]);
        Bd.remove(1);
        if (g != one && S != zeropoly)
          M.appendrow(S.coeffs);
      }
      while (Vd.length() > 0) {
        v = Vd[1];
        Vd.remove(1);
        M.appendrow(v.coeffs);
      }
      M.Gauss(start);

      for (i = start; i <= M.getheight(); i++) {
        nr = homopoly<numeric>(d, M[i]);
        s++;
        G.append(nr);
        if (opt_KR)
          Update(B, Bast, G);
        else if (opt_GM)
          pair_optimize(B, G);
        else
          pair_update(B, G);
      }
    }
    d++;
  }

  return G;
}

/***************************************************************************/
/* Buchberger algorithm for homogeneous ideals using Gauss with statistics */
/***************************************************************************/

template<class numeric> list<homopoly<numeric> > StatGBBuchbergerGauss(const homoideal<numeric> &I) {
  homopoly        <numeric>   nr, S, v, zeropoly;
  list<homopairset<numeric> > A, B, Bast, Bastd, Bd;
  list<homopoly   <numeric> > G, V = I.gens, Vd;
  monomial                    g, one;
  sparsematrix    <numeric>   M;
  double        sp, sp_min = 100, sp_max = 0;
  string        comp_time;
  unsigned long d, dmin, i, j, s = 0, start,
                mat_height = 0, mat_width  = 0,
                pair_all   = 0, pair_count = 0, pair_used = 0,
                sp_mindeg  = 0, sp_maxdeg = 0;

  if (V.length() == 0)
    return G;

  d    = V[1].deg();
  dmin = d;
  for (i = 2; i <= V.length(); i++) {
    d    = V[i].deg();
    dmin = min(dmin, d);
  }
  d = dmin;

  while (B.length() > 0 || V.length() > 0) {
    for (i = 1; i <= B.length();)
      if (B[i].deg() == d) {
        Bd.append(B[i]);
        B.remove(i);
      }
      else
        i++;
    if (opt_NSS)
      PairSort(Bd);
    for (i = 1; i <= V.length();)
      if (V[i].deg() == d) {
        Vd.append(V[i]);
        V.remove(i);
      }
      else
        i++;
    if (opt_KR) {
      for (i = 1; i <= Bast.length();)
        if (Bast[i].deg() == d) {
          Bastd.append(Bast[i]);
          Bast.remove(i);
        }
        else
          i++;
      MinPairs(A, Bd, Bastd, G);
    }
    M  = MatrixUpdate(M, d, G);

    if (Bd.length() > 0 || Vd.length() > 0) {
      start = M.getheight() + 1;
      while (Bd.length() > 0) {
        pair_used++;
        g = Bd[1].g;
        S = Bd[1].SVector;
        if (opt_KR)
          A.append(Bd[1]);
        Bd.remove(1);
        if (g != one && S != zeropoly)
          M.appendrow(S.coeffs);
      }
      while (Vd.length() > 0) {
        v = Vd[1];
        Vd.remove(1);
        M.appendrow(v.coeffs);
      }
      mat_height = M.getheight  ();
      mat_width  = M.getwidth   ();
      sp         = M.getsparsity();
      M.Gauss(start);
      if (sp < sp_min) {
        sp_min    = sp;
        sp_mindeg = d ;
      }
      if (sp > sp_max) {
        sp_max    = sp;
        sp_maxdeg = d ;
      }

      for (i = start; i <= M.getheight(); i++) {
        nr = homopoly<numeric>(d, M[i]);
        s++;
        G.append(nr);
        pair_all += s - 1;
        j         = B.length();
        if (opt_KR)
          Update(B, Bast, G);
        else if (opt_GM)
          pair_optimize(B, G);
        else
          pair_update(B, G);
        pair_count += B.length() - j;
      }
    }
    d++;
  }

  comp_time = FormatTime(timer('e'));
  cout << endl
       << replicate('', 70)                                                                                    << endl
       << "algorithm:              BuchbergerGauss"                                                             << endl
       << "optimizations:          " << GetOptString()                                                          << endl
       << replicate('', 79)                                                                                    << endl
       << "computed degrees: min:  " << dmin                                                                    << endl
       << "                  max:  " << (d - 1)                                                                 << endl
       << "number of all    pairs: " << pair_all                                                                << endl
       << "number of listed pairs: " << pair_count                                                              << endl
       << "number of used   pairs: " << pair_used                                                               << endl
       << "original generators:    " << I.gens.length()                                                         << endl
       << "GB       generators:    " << G.length()                                                              << endl
       << "new      generators:    " << (set<homopoly<numeric> >(G) - set<homopoly<numeric> >(I.gens)).length() << endl
       << "minimum sparsity:       " << sp_min << "% (degree " << sp_mindeg << ")"                              << endl
       << "maximum sparsity:       " << sp_max << "% (degree " << sp_maxdeg << ")"                              << endl
       << "size of last matrix:    " << mat_height << " x " << mat_width                                        << endl
       << "computation time:       " << comp_time                                                               << endl
       << "GB verification:        " << (isGB(G) ? "ok" : "failure")                                            << endl
       << replicate('', 79)                                                                                    << endl
       << endl;

  return G;
}

/*********************************************************************/
/* Buchberger algorithm for homogeneous ideals using multi reduction */
/*********************************************************************/

template<class numeric> list<homopoly<numeric> > GBBuchbergerMulti(const homoideal<numeric> &I) {
  homopoly        <numeric>   zeropoly;
  list<homopairset<numeric> > A, B, Bast, Bastd, Bd;
  list<homopoly   <numeric> > F, G, V = I.gens;
  monomial                    one;
  unsigned long d, i, s = 0;

  if (V.length() == 0)
    return G;

  while (B.length() > 0 || V.length() > 0) {
    if (B.length() == 0)
      d = MinDeg(V);
    else if (V.length() == 0)
      d = MinDeg(B);
    else
      d = min(MinDeg(B), MinDeg(V));
    for (i = 1; i <= B.length();)
      if (B[i].deg() == d) {
        Bd.append(B[i]);
        B.remove(i);
      }
      else
        i++;
    if (opt_KR) {
      for (i = 1; i <= Bast.length();)
        if (Bast[i].deg() == d) {
          Bastd.append(Bast[i]);
          Bast.remove(i);
        }
        else
          i++;
      MinPairs(A, Bd, Bastd, G);
      A += Bd;
    }
    F.clear();
    for (i = 1; i <= Bd.length(); i++)
      if (Bd[i].g != one && Bd[i].SVector != zeropoly)
        F.append(Bd[i].SVector);
    Bd.clear();
    for (i = 1; i <= V.length();)
      if (V[i].deg() == d) {
        F.append(V[i]);
        V.remove(i);
      }
      else
        i++;
    MR(F, G, d);
    for (i = 1; i <= F.length(); i++) {
      s++;
      G.append(F[i]);
      if (opt_KR)
        Update(B, Bast, G);
      else if (opt_GM)
        pair_optimize(B, G);
      else
        pair_update(B, G);
    }
  }

  return G;
}

/*************************************************************************************/
/* Buchberger algorithm for homogeneous ideals using multi reduction with statistics */
/*************************************************************************************/

template<class numeric> list<homopoly<numeric> > StatGBBuchbergerMulti(const homoideal<numeric> &I) {
  homopoly        <numeric>   zeropoly;
  list<homopairset<numeric> > A, B, Bast, Bastd, Bd;
  list<homopoly   <numeric> > F, G, V = I.gens;
  monomial                    one;
  string        comp_time;
  unsigned long d, dmin, i, j, s = 0,
                nr_count = 0, nr_steps   = 0, nr_zero   = 0,
                pair_all = 0, pair_count = 0, pair_used = 0;

  if (V.length() == 0)
    return G;

  timer('s');
  dmin = V[1].deg();
  while (B.length() > 0 || V.length() > 0) {
    if (B.length() == 0)
      d = MinDeg(V);
    else if (V.length() == 0)
      d = MinDeg(B);
    else
      d = min(MinDeg(B), MinDeg(V));
    dmin = min(dmin, d);
if (dooutput) {
cerr << endl;
cerr << "d = " << d << endl;
cerr << "G = " << G << endl;
cerr << "B = " << B << endl;
cerr << "V = " << V << endl;
cout << endl;
cout << "d = " << d << endl;
cout << "G = " << G << endl;
cout << "B = " << B << endl;
cout << "V = " << V << endl;
}
    for (i = 1; i <= B.length();)
      if (B[i].deg() == d) {
        Bd.append(B[i]);
        B.remove(i);
      }
      else
        i++;
if (dooutput) {
cerr << "B" << d << " = " << Bd << endl;
cout << "B" << d << " = " << Bd << endl;
}
    if (opt_KR) {
      for (i = 1; i <= Bast.length();)
        if (Bast[i].deg() == d) {
          Bastd.append(Bast[i]);
          Bast.remove(i);
        }
        else
          i++;
      MinPairs(A, Bd, Bastd, G);
      A += Bd;
    }
    F.clear();
    pair_used += Bd.length();
    for (i = 1; i <= Bd.length(); i++)
      if (Bd[i].g != one && Bd[i].SVector != zeropoly)
        F.append(Bd[i].SVector);
    Bd.clear();
    for (i = 1; i <= V.length();)
      if (V[i].deg() == d) {
        F.append(V[i]);
        V.remove(i);
      }
      else
        i++;
    nr_count++;
    i = F.length();
    MR(F, G, d, nr_steps);
    nr_zero += i - F.length();
    for (i = 1; i <= F.length(); i++) {
      s++;
      G.append(F[i]);
      pair_all += s - 1;
      j = B.length();
      if (opt_KR)
        Update(B, Bast, G);
      else if (opt_GM)
        pair_optimize(B, G);
      else
        pair_update(B, G);
      pair_count += B.length() - j;
    }
  }

  comp_time = FormatTime(timer('e'));
  cout << endl
       << replicate('', 79)                                                                                    << endl
       << "algorithm:              BuchbergerMulti"                                                             << endl
       << "optimizations:          " << GetOptString()                                                          << endl
       << replicate('', 79)                                                                                    << endl
       << "computed degrees: min:  " << dmin                                                                    << endl
       << "                  max:  " << d                                                                       << endl
       << "number of all    pairs: " << pair_all                                                                << endl
       << "number of listed pairs: " << pair_count                                                              << endl
       << "number of used   pairs: " << pair_used                                                               << endl
       << "multi reductions:       " << nr_count << " (" << nr_zero << " vanished)"                             << endl
       << "reduction steps:        " << nr_steps                                                                << endl
       << "original generators:    " << I.gens.length()                                                         << endl
       << "GB       generators:    " << G.length()                                                              << endl
       << "new      generators:    " << (set<homopoly<numeric> >(G) - set<homopoly<numeric> >(I.gens)).length() << endl
       << "computation time:       " << comp_time                                                               << endl
//       << "GB verification:        " << (isGB(G) ? "ok" : "failure")                                            << endl
       << replicate('', 79)                                                                                    << endl
       << endl;

  return G;
}

/*****************************/
/* F4 for homogeneous ideals */
/*****************************/

template<class numeric> list<homopoly<numeric> > F4Homo(const homoideal<numeric> &I) {
  list<homopairset<numeric> >     A, B, Bast, Bastd, Bd;
  list<homopoly   <numeric> >     FF, G, L, V = I.gens;
  list<list<homopoly<numeric> > > F, Ftil;
  unsigned long d, i, s = 0;

  if (V.length() == 0)
    return G;

  while (B.length() > 0 || V.length() > 0) {
    if (B.length() == 0)
      d = MinDeg(V);
    else if (V.length() == 0)
      d = MinDeg(B);
    else
      d = min(MinDeg(B), MinDeg(V));
    for (i = 1; i <= B.length();)
      if (B[i].deg() == d) {
        Bd.append(B[i]);
        B.remove(i);
      }
      else
        i++;
    FF.clear();
    for (i = 1; i <= V.length();)
      if (V[i].deg() == d) {
        FF.append(V[i]);
        V.remove(i);
      }
      else
        i++;
    if (opt_KR) {
      for (i = 1; i <= Bast.length();)
        if (Bast[i].deg() == d) {
          Bastd.append(Bast[i]);
          Bast.remove(i);
        }
        else
          i++;
      MinPairs(A, Bd, Bastd, G);
      A += Bd;
    }
    while (Bd.length() > 0 || FF.length() > 0) {
      if (FF.length() == 0) {
        L.clear();
        for (i = 1; i <= Bd.length(); i++) {
          L.append(Simp(Bd[i].tij, G[Bd[i].val1], F, Ftil));
          L.append(Simp(Bd[i].tji, G[Bd[i].val2], F, Ftil));
        }
        FF = FR(L, G, F, Ftil);
        Bd.clear();
      }
      for (i = 1; i <= FF.length(); i++) {
        s++;
        G.append(FF[i]);
        if (opt_KR)
          Update(B, Bast, G);
        else if (opt_GM)
          pair_optimize(B, G);
        else
          pair_update(B, G);
      }
      if (opt_KR) {
        for (i = 1; i <= Bast.length();)
          if (Bast[i].deg() == d) {
            Bastd.append(Bast[i]);
            Bast.remove(i);
          }
          else
            i++;
        MinPairs(A, Bd, Bastd, G);
        A += Bd;
      }
      FF.clear();
    }
  }

  return G;
}

/*********************************************/
/* F4 for homogeneous ideals with statistics */
/*********************************************/

template<class numeric> list<homopoly<numeric> > StatF4Homo(const homoideal<numeric> &I) {
  list<homopairset<numeric> >     A, B, Bast, Bastd, Bd;
  list<homopoly   <numeric> >     FF, G, L, V = I.gens;
  list<list<homopoly<numeric> > > F, Ftil;
  string        comp_time;
  unsigned long d, dmin, i, j, s = 0,
                nr_count = 0, nr_simp    = 0, nr_steps  = 0,
                pair_all = 0, pair_count = 0, pair_used = 0;

  if (V.length() == 0)
    return G;

  timer('s');
  dmin = V[1].deg();
  while (B.length() > 0 || V.length() > 0) {
    if (B.length() == 0)
      d = MinDeg(V);
    else if (V.length() == 0)
      d = MinDeg(B);
    else
      d = min(MinDeg(B), MinDeg(V));
if (dooutput) {
cerr << endl;
cerr << "d = " << d << endl;
cerr << "G = " << G << endl;
cerr << "B = " << B << endl;
cerr << "V = " << V << endl;
}
    dmin = min(dmin, d);
    for (i = 1; i <= B.length();)
      if (B[i].deg() == d) {
        Bd.append(B[i]);
        B.remove(i);
      }
      else
        i++;
if (dooutput)
cerr << "B" << d << " = " << Bd << endl;
    FF.clear();
    for (i = 1; i <= V.length();)
      if (V[i].deg() == d) {
        FF.append(V[i]);
        V.remove(i);
      }
      else
        i++;
    if (opt_KR) {
      for (i = 1; i <= Bast.length();)
        if (Bast[i].deg() == d) {
          Bastd.append(Bast[i]);
          Bast.remove(i);
        }
        else
          i++;
      MinPairs(A, Bd, Bastd, G);
      A += Bd;
    }
    while (Bd.length() > 0 || FF.length() > 0) {
      if (FF.length() == 0) {
        L.clear();
        for (i = 1; i <= Bd.length(); i++) {
          L.append(Simp(Bd[i].tij, G[Bd[i].val1], F, Ftil, nr_simp));
          L.append(Simp(Bd[i].tji, G[Bd[i].val2], F, Ftil, nr_simp));
        }
        nr_count++;
        FF = FR(L, G, F, Ftil, nr_steps, nr_simp);
        pair_used += Bd.length();
        Bd.clear();
      }
      for (i = 1; i <= FF.length(); i++) {
        s++;
        G.append(FF[i]);
        pair_all += s - 1;
        j = B.length();
        if (opt_KR)
          Update(B, Bast, G);
        else if (opt_GM)
          pair_optimize(B, G);
        else
          pair_update(B, G);
        pair_count += B.length() - j;
      }
      if (opt_KR) {
        for (i = 1; i <= Bast.length();)
          if (Bast[i].deg() == d) {
            Bastd.append(Bast[i]);
            Bast.remove(i);
          }
          else
            i++;
        MinPairs(A, Bd, Bastd, G);
        A += Bd;
      }
      FF.clear();
    }
  }

  comp_time = FormatTime(timer('e'));
  cout << endl
       << replicate('', 79)                                                                                    << endl
       << "algorithm:              F4"                                                                          << endl
       << "optimizations:          " << GetOptString()                                                          << endl
       << replicate('', 79)                                                                                    << endl
       << "computed degrees: min:  " << dmin                                                                    << endl
       << "                  max:  " << d                                                                       << endl
       << "number of all    pairs: " << pair_all                                                                << endl
       << "number of listed pairs: " << pair_count                                                              << endl
       << "number of used   pairs: " << pair_used                                                               << endl
       << "F-reductions:           " << nr_count                                                                << endl
       << "Gauss reduction steps:  " << nr_steps                                                                << endl
       << "simplification  steps:  " << nr_simp                                                                 << endl
       << "original generators:    " << I.gens.length()                                                         << endl
       << "GB       generators:    " << G.length()                                                              << endl
       << "new      generators:    " << (set<homopoly<numeric> >(G) - set<homopoly<numeric> >(I.gens)).length() << endl
       << "computation time:       " << comp_time                                                               << endl
//       << "GB verification:        " << (isGB(G) ? "ok" : "failure")                                            << endl
       << replicate('', 79)                                                                                    << endl
       << endl;

  return G;
}


/************************/
/* other used functions */
/************************/

// get minimum degree

template<class T> unsigned long MinDeg(list<T> &L) {
  unsigned long d, i, l = L.length();
  if (l == 0)
    return 0;
  d = L[1].deg();
  for (i = 2; i <= l; i++)
    d = min(d, (unsigned long) L[i].deg());
  return d;
}


// normal selection strategy

template<class pair> void PairSort(list<pair> &B) {
  list<pair> A = B;
  monomial m;
  Z d;
  unsigned long i, j, k, l = B.length(), pos;

  // maybe implement quick sort(s) here some day... hehe

  // degree compatible orderings:
  if (TermOrd == DegLex || TermOrd == DegRevLex) {
    for (i = 1; i <= l; i++) {
      k = l - i + 1;
      m = A[1].l;
      pos = 1;
      for (j = 2; j <= k; j++)
        if (A[j].l < m) {
          m = A[j].l;
          pos = j;
        }
      B[i] = A[pos];
      A.remove(pos);
    }
  }

  // not degree compatible orderings:
  else {
    for (i = 1; i <= l; i++) {
      k = l - i + 1;
      d = A[1].l.deg();
      pos = 1;
      for (j = 2; j <= k; j++)
        if (A[j].l.deg() < d) {
          d = A[j].l.deg();
          pos = j;
        }
      B[i] = A[pos];
      A.remove(pos);
    }
  }
}


// Update without optimization

template<class numeric> void pair_update(list<pairset<numeric> > &B, list<poly<numeric> > &G) {
  unsigned long i, s = G.length();
  for (i = 1; i < s; i++)
    B.append(pairset<numeric>(i, s, &G[i], &G[s]));
}

template<class numeric> void pair_update(list<homopairset<numeric> > &B, list<homopoly<numeric> > &G) {
  unsigned long i, s = G.length();
  for (i = 1; i < s; i++)
    B.append(homopairset<numeric>(i, s, &G[i], &G[s]));
}


// Gebauer Mller optimizations

template<class numeric> void pair_optimize(list<pairset<numeric> > &B, list<poly<numeric> > &G) {
  unsigned long i, j, k, s = G.length();
  monomial l, one, ti, tij, tj, tji;
  list<pairset<numeric> > C;
  bool rem;

  for (i = 1; i < s; i++)
    C.append(pairset<numeric>(i, s, &G[i], &G[s]));

  if (C.length() > 1)
    for (j = C.length(); j > 1; j--)
      for (i = 1; i < j; i++)
        if (C[i].tji.devides(C[j].tji)) {
          C.remove(j);
          break;
        }

  if (C.length() > 1)
    for (i = C.length() - 1; i > 0; i--)
      for (j = i + 1; j <= C.length(); j++)
        if (C[j].tji.devides(C[i].tji) && C[j].tji != C[i].tji) {
          C.remove(i);
          break;
        }

  if (C.length() > 1)
    for (i = B.length(); i > 0; i--) {
      rem = false;
      for (j = 1; j <= C.length(); j++)
        if (C[j].val1 == B[i].val1) {
          ti = G[C[j].val1].LT();
          for (k = j + 1; k <= C.length(); k++)
            if (C[k].val1 == B[i].val2) {
              tj  = G[C[k].val1].LT();
              l   = lcm(ti, tj);
              tij = l / ti;
              tji = l / tj;
              if (C[j].tij.devides(tij) && C[j].tij != tij && C[k].tij.devides(tji) && C[k].tij != tji) {
                rem = true;
                break;
              }
            }
          if (rem) {
            B.remove(i);
            break;
          }
        }

    }

  B += C;
}

template<class numeric> void pair_optimize(list<homopairset<numeric> > &B, list<homopoly<numeric> > &G) {
  unsigned long i, j, k, s = G.length();
  monomial l, one, ti, tij, tj, tji;
  list<homopairset<numeric> > C;
  bool rem;

  for (i = 1; i < s; i++)
    C.append(homopairset<numeric>(i, s, &G[i], &G[s]));

  if (C.length() > 1)
    for (j = C.length(); j > 1; j--)
      for (i = 1; i < j; i++)
        if (C[i].tji.devides(C[j].tji)) {
          C.remove(j);
          break;
        }

  if (C.length() > 1)
    for (i = C.length() - 1; i > 0; i--)
      for (j = i + 1; j <= C.length(); j++)
        if (C[j].tji.devides(C[i].tji) && C[j].tji != C[i].tji) {
          C.remove(i);
          break;
        }

  if (C.length() > 1)
    for (i = B.length(); i > 0; i--) {
      rem = false;
      for (j = 1; j <= C.length(); j++)
        if (C[j].val1 == B[i].val1) {
          ti = G[C[j].val1].LT();
          for (k = j + 1; k <= C.length(); k++)
            if (C[k].val1 == B[i].val2) {
              tj  = G[C[k].val1].LT();
              l   = lcm(ti, tj);
              tij = l / ti;
              tji = l / tj;
              if (C[j].tij.devides(tij) && C[j].tij != tij && C[k].tij.devides(tji) && C[k].tij != tji) {
                rem = true;
                break;
              }
            }
          if (rem) {
            B.remove(i);
            break;
          }
        }

    }

  B += C;
}


// Kreuzer Robbiano optimizations

template<class numeric> void Update(list<homopairset<numeric> > &B, list<homopairset<numeric> > &Bast, list<homopoly<numeric> > &G) {
  unsigned long i, j, k, s = G.length();
  monomial l, ti, tij, tj, tji, one;
  list<homopairset<numeric> > C;
  bool add;

  // U1)
  for (i = 1; i < s; i++)
    C.append(homopairset<numeric>(i, s, &G[i], &G[s]));

  // U2)
  if (C.length() > 1)
    for (j = C.length(); j > 1; j--)
      for (i = 1; i < j; i++)
        if (C[i].tji.devides(C[j].tji)) {
          C.remove(j);
          break;
        }

  // U3)
  if (C.length() > 1)
    for (i = C.length() - 1; i > 0; i--)
      for (j = i + 1; j <= C.length(); j++)
        if (C[j].tji.devides(C[i].tji) && C[j].tji != C[i].tji) {
          C.remove(i);
          break;
        }

  // U4)
  if (C.length() > 1)
    for (i = 1; i < C.length(); i++)
      for (j = i + 1; j <= C.length(); j++)
        if (gcd(C[i].tij, C[j].tij) == one) {
          add = true;
          for (k = 1; k <= Bast.length(); k++)
            if (Bast[k].val1 == C[i].val1 && Bast[k].val2 == C[j].val1) {
              add = false;
              break;
            }
          if (add)
            Bast.append(homopairset<numeric>(C[i].val1, C[j].val1, &G[C[i].val1], &G[C[j].val1]));
        }

  // U5)
  B += C;
}

template<class numeric> void MinPairs(list<homopairset<numeric> > &A, list<homopairset<numeric> > &Bd, list<homopairset<numeric> > &Bastd, list<homopoly<numeric> > &G) {
  homopairset<numeric> pair1, pair2, pair3;
  unsigned long i, j, k, l;
  bool     done;
  monomial lc, tk, tl, tlk, one;

  while (Bastd.length() > 0) {
    pair1 = Bastd[1];
    Bastd.remove(1);
    done = false;

    for (i = 1; i <= A.length(); i++) {
      pair2 = A[i];
      if (pair1.val2 == pair2.val2 && pair1.tji == pair2.tji) {
        done = true;
        break;
      }
    }

    if (!done)
      for (i = 1; i <= Bd.length(); i++) {
        pair2 = Bd[i];
        if (pair1.val2 == pair2.val2 && pair1.tji == pair2.tji) {
          Bd.remove(i);
          A.append(pair2);
          done = true;
          break;
        }
      }

    while (!done) {
      for (i = 1; i <= A.length(); i++) {
        pair2 = A[i];
        if (pair1.val2 == pair2.val2 && pair2.tji.devides(pair1.tji))
          break;
      }
      if (gcd(pair1.tij, pair2.tij) == one) {
        done  = true;
        break;
      }
      if (pair1.val1 < pair2.val1) {
        k = pair1.val1;
        l = pair2.val1;
      }
      else {
        k = pair2.val1;
        l = pair1.val1;
      }
      tk  = G[k].LT();
      tl  = G[l].LT();
      lc  = lcm(tk, tl);
      tlk = lc / tl;
      for (j = 1; j <= A.length(); j++) {
        pair3 = A[j];
        if (pair3.val2 == l && pair3.tji == tlk) {
          done  = true;
          break;
        }
      }
      if (done)
        break;
      for (j = 1; j <= Bd.length(); j++) {
        pair3 = Bd[j];
        if (pair3.val2 == l && pair3.tji == tlk) {
          Bd.remove(j);
          A.append(pair3);
          done = true;
          break;
        }
      }
      if (done)
        break;
        for (j = 1; j <= Bastd.length(); j++) {
          pair3 = Bd[j];
          if (pair3.val1 == k && pair3.val2 == l) {
            Bastd.remove(j);
            pair1 = homopairset<numeric>(k, l, &G[k], &G[l]);
            break;
          }
        }
    }
  }
}


// check if Grbner basis

template<class polynomial> bool isGB(list<polynomial> &G) {
  unsigned long i, j;
  monomial l, tij, tji;
  polynomial S, zero;
  for (i = 1; i < G.length(); i++)
    for (j = i + 1; j <= G.length(); j++) {
      l   = lcm(G[i].LT(), G[j].LT());
      tij = l / G[i].LT();
      tji = l / G[j].LT();
      S   = tij * G[i] / G[i].LC();
      S  -= tji * G[j] / G[j].LC();
      if (S.NR(G) != zero)
        return false;
    }
  return true;
}


// Update matrix in preparation for Gauss

template<class numeric> sparsematrix<numeric> MatrixUpdate(const sparsematrix<numeric> &Mold, const unsigned long &d, list<homopoly<numeric> > &G) {
  unsigned long comp, first, i, j, move, offset, range,
                n         = IndetList.entries      ,
                oldheight = Mold.getheight()       ,
                oldwidth  = Mold.getwidth()        ,
                width     = binomial(n + d - 1, d) ;
  sparseentry <numeric> *entry;
  sparsevector<numeric> oldrow, row(width);
  sparsematrix<numeric> M;
  list<bool> filled(width, false);
  monomial m;

  // transform M from d - 1 to d
  switch (TermOrd) {
    case Lex      :
    case DegLex   : for (i = 1; i <= oldheight; i++) {
                      oldrow = Mold[i];
                      if (!oldrow.first)
                        continue;
                      first  = oldrow.first->position;
                      offset = 0;
                      for (comp = 1; comp <= n; comp++) {
                        range = binomial(n + d - comp - 1, d - 1);
                        move  = oldwidth - range;
                        if (first <= move)
                          break;
                        move  = offset - move;
                        entry = oldrow.first;
                        filled[entry->position + move] = true;
                        row.clear(width);
                        while (entry) {
                          row.append(entry->position + move, entry->value);
                          entry = entry->next;
                        }
                        M.appendrow(row);
                        offset += range;
                      }
                    }
                    break;
    case DegRevLex: for (i = 1; i <= oldheight; i++) {
                      oldrow = Mold[i];
                      if (!oldrow.first)
                        continue;
                      first  = oldrow.first->position;
                      offset = 0;
                      for (comp = n; comp > 0; comp--) {
                        range = binomial(n + d - comp - 1, d - 1);
                        if (first <= range) {
                          entry = oldrow.first;
                          filled[entry->position + offset] = true;
                          row.clear(width);
                          while (entry) {
                            row.append(entry->position + offset, entry->value);
                            entry = entry->next;
                          }
                          M.appendrow(row);
                        }
                        offset += range;
                      }
                    }
                    break;
    case Elim     :
    default       :;
  }

  // find new lines
  for (i = 1; i <= width; i++)
    if (!filled[i]) {
      m = indets.GetHomoMono(d, i);
      for (j = 1; j <= G.length(); j++)
        if (G[j].LT().devides(m)) {
          row = (G[j] * (m / G[j].LT())).coeffs;
          M.appendrow(row);
        }
    }

  M.RowEchelon();
  return M;
}


// multi reduction

template<class numeric> void MR(list<homopoly<numeric> > &F, list<homopoly<numeric> > &G, const unsigned long &d) {
  unsigned long i, j, m = F.length(), minpos, s = G.length();
  list<sparseentry<numeric>*> actual(m);
  monomial redterm;
  homopoly<numeric> diff, zeropoly;
  numeric coeff;
  bool reduced;

  if (m == 0 || s == 0)
    return;

  // reducing F with polynomials in G
  for (i = 1; i <= m; i++)
    if (F[i].coeffs.first)
      actual[i] = F[i].coeffs.first;
  while (true) {
    minpos = -1lu;
    for (i = 1; i <= m; i++)
      if (actual[i])
        minpos = min(minpos, actual[i]->position);
    if (minpos == -1lu)
      break;
    redterm = indets.GetHomoMono(d, minpos);
    for (i = 1; i <= s; i++)
      if (G[i].LT().devides(redterm)) {
        redterm /= G[i].LT();
        diff = redterm * G[i];
        break;
      }
    if (i <= s) {
      for (i = 1; i <= m; i++)
        if (actual[i] != NULL && actual[i]->position == minpos) {
          coeff = actual[i]->value;
          F[i] -= coeff / diff.LC() * diff;
          actual[i] = F[i].coeffs.first;
          while (actual[i]->position < minpos && actual[i] != NULL)
            actual[i] = actual[i]->next;
        }
    }
    else
      for (i = 1; i <= m; i++)
        if (actual[i] != NULL && actual[i]->position == minpos)
          actual[i] = actual[i]->next;
  }

  // ordering F, head terms decreasing
  PolySort(F, 1);

  // reducing F with polynomials in F
  for (i = 1; i < F.length(); i++) {
    reduced = false;
    for (j = i + 1; j <= F.length() && F[j].LT() == F[i].LT(); j++)
      if (F[i].LT() == F[j].LT()) {
        F[j] -= F[j].LC() / F[i].LC() * F[i];
        reduced = true;
        if (F[j] == zeropoly) {
          F.remove(j);
          j--;
        }
      }
    if (reduced)
      PolySort(F, i + 1);
  }
}

template<class numeric> void MR(list<homopoly<numeric> > &F, list<homopoly<numeric> > &G, const unsigned long &d, unsigned long &count) {
  unsigned long i, j, m = F.length(), minpos, s = G.length();
  list<sparseentry<numeric>*> actual(m);
  monomial redterm;
  homopoly<numeric> diff, zeropoly;
  numeric coeff;
  bool reduced;

  if (m == 0 || s == 0)
    return;

  // reducing F with polynomials in G
  for (i = 1; i <= m; i++)
    if (F[i].coeffs.first)
      actual[i] = F[i].coeffs.first;
  while (true) {
    minpos = -1lu;
    for (i = 1; i <= m; i++)
      if (actual[i])
        minpos = min(minpos, actual[i]->position);
    if (minpos == -1lu)
      break;
    redterm = indets.GetHomoMono(d, minpos);
    for (i = 1; i <= s; i++)
      if (G[i].LT().devides(redterm)) {
        redterm /= G[i].LT();
        diff = redterm * G[i];
        break;
      }
    if (i <= s) {
      for (i = 1; i <= m; i++)
        if (actual[i] != NULL && actual[i]->position == minpos) {
          coeff = actual[i]->value;
          F[i] -= coeff / diff.LC() * diff;
          count++;
          actual[i] = F[i].coeffs.first;
          while (actual[i]->position < minpos && actual[i] != NULL)
            actual[i] = actual[i]->next;
        }
    }
    else
      for (i = 1; i <= m; i++)
        if (actual[i] != NULL && actual[i]->position == minpos)
          actual[i] = actual[i]->next;
  }

  // reducing F with polynomials in F
  PolySort(F, 1);
  for (i = 1; i < F.length(); i++) {
    reduced = false;
    for (j = i + 1; j <= F.length() && F[j].LT() == F[i].LT(); j++)
      if (F[i].LT() == F[j].LT()) {
        diff = F[j].LC() / F[i].LC() * F[i];
        F[j] -= diff;
        reduced = true;
        count++;
        if (F[j] == zeropoly) {
          F.remove(j);
          j--;
        }
      }
    if (reduced)
      PolySort(F, i + 1);
  }
}


// F4 functions

template<class numeric> list<homopoly<numeric> > FR(list<homopoly<numeric> > L, list<homopoly<numeric> > &G, list<list<homopoly<numeric> > > &F, list<list<homopoly<numeric> > > &Ftil) {
  list<homopoly<numeric> > LL;
  set<homopoly<numeric> > LLL;
  set<monomial> D, S, Su;
  homopoly<numeric> f;
  monomial m, t;
  unsigned long i, s = G.length();
  for (i = 1; i <= L.length(); i++) {
    m = L[i].LT();
    if (!D.contains(m))
      D.append(m);
    S += L[i].supp();
  }
  S -= D;
  while (S.length() > 0) {
    m = S[1];
    S.erase(m);
    D.append(m);
    for (i = 1; i <= s; i++) {
      t = G[i].LT();
      if (t.devides(m)) {
        f = Simp(m / t, G[i], F, Ftil);
        L.append(f);
        Su  = f.supp();
        Su -= D;
        S  += Su;
        break;
      }
    }
  }
  F.append(L);
  LL = interreduce(L);
  Ftil.append(LL);
  LLL = LL;
  LLL -= L;
  return LLL;
}

template<class numeric> list<homopoly<numeric> > FR(list<homopoly<numeric> > L, list<homopoly<numeric> > &G, list<list<homopoly<numeric> > > &F, list<list<homopoly<numeric> > > &Ftil, unsigned long &count1, unsigned long &count2) {
  list<homopoly<numeric> > LL;
  set<homopoly<numeric> > LLL;
  set<monomial> D, S, Su;
  homopoly<numeric> f;
  monomial m, t;
  unsigned long i, s = G.length();
  for (i = 1; i <= L.length(); i++) {
    m = L[i].LT();
    if (!D.contains(m))
      D.append(m);
    S += L[i].supp();
  }
  S -= D;
  while (S.length() > 0) {
    m = S[1];
    S.erase(m);
    D.append(m);
    for (i = 1; i <= s; i++) {
      t = G[i].LT();
      if (t.devides(m)) {
        f = Simp(m / t, G[i], F, Ftil, count2);
        L.append(f);
        Su  = f.supp();
        Su -= D;
        S  += Su;
        break;
      }
    }
  }
  F.append(L);
  LL = interreduce(L, count1);
  Ftil.append(LL);
  LLL = LL;
  LLL -= L;
  return LLL;
}

template<class numeric> list<homopoly<numeric> > interreduce(list<homopoly<numeric> > L) {
  list<homopoly<numeric> > LL;
  homopoly<numeric> zeropoly;
  unsigned long a = L.length(), i, j, l = L.length(), ll, pos, poslen;
  list<unsigned long> active(a), leaders;
  sparseentry<numeric>* entry;

  for (i = 1; i <= a; i++)
    active[i] = i;

  for (; a > 0; a--) {
    for (i = a; i > 0; i--)
      if (L[active[i]] == zeropoly) {
        active.remove(i);
        a--;
      }
    if (a == 0)
      break;
    pos = L[active[1]].coeffs.first->position;
    for (i = 2; i <= a; i++)
      pos = min(pos, L[active[i]].coeffs.first->position);
    leaders.clear();
    for (i = 1; i <= a; i++)
      if (pos == L[active[i]].coeffs.first->position)
        leaders.append(i);
    ll = leaders.length();
    pos = leaders[1];
    poslen = L[active[leaders[1]]].coeffs.entries();
    for (i = 2; i <= ll; i++)
      if (L[active[leaders[i]]].coeffs.entries() < poslen) {
        pos = leaders[i];
        poslen = L[active[leaders[i]]].coeffs.entries();
      }
    L[active[pos]] /= L[active[pos]].LC();
    for (i = 1; i <= ll; i++)
      if (pos != leaders[i])
        L[active[leaders[i]]] -= L[active[leaders[i]]].LC() * L[active[pos]];
    LL.append(L[active[pos]]);
    active.remove(pos);
  }
  l = LL.length();
  for (i = l; i > 1; i--) {
    pos = LL[i].coeffs.first->position;
    for (j = 1; j < i; j++) {
      entry = LL[j].coeffs.first;
      while (entry != NULL && entry->position < pos)
        entry = entry->next;
      if (entry != NULL && entry->position == pos)
        LL[j] -= entry->value * LL[i];
    }
  }
  return LL;
}

template<class numeric> list<homopoly<numeric> > interreduce(list<homopoly<numeric> > L, unsigned long &count) {
  list<homopoly<numeric> > LL;
  homopoly<numeric> zeropoly;
  unsigned long a = L.length(), i, j, l = L.length(), ll, pos, poslen;
  list<unsigned long> active(a), leaders;
  sparseentry<numeric>* entry;

  for (i = 1; i <= a; i++)
    active[i] = i;

  for (; a > 0; a--) {
    for (i = a; i > 0; i--)
      if (L[active[i]] == zeropoly) {
        active.remove(i);
        a--;
      }
    if (a == 0)
      break;
    pos = L[active[1]].coeffs.first->position;
    for (i = 2; i <= a; i++)
      pos = min(pos, L[active[i]].coeffs.first->position);
    leaders.clear();
    for (i = 1; i <= a; i++)
      if (pos == L[active[i]].coeffs.first->position)
        leaders.append(i);
    ll = leaders.length();
    pos = leaders[1];
    poslen = L[active[leaders[1]]].coeffs.entries();
    for (i = 2; i <= ll; i++)
      if (L[active[leaders[i]]].coeffs.entries() < poslen) {
        pos = leaders[i];
        poslen = L[active[leaders[i]]].coeffs.entries();
      }
    L[active[pos]] /= L[active[pos]].LC();
    for (i = 1; i <= ll; i++)
      if (pos != leaders[i]) {
        L[active[leaders[i]]] -= L[active[leaders[i]]].LC() * L[active[pos]];
        count++;
      }
    LL.append(L[active[pos]]);
    active.remove(pos);
  }
  l = LL.length();
  for (i = l; i > 1; i--) {
    pos = LL[i].coeffs.first->position;
    for (j = 1; j < i; j++) {
      entry = LL[j].coeffs.first;
      while (entry != NULL && entry->position < pos)
        entry = entry->next;
      if (entry != NULL && entry->position == pos) {
        LL[j] -= entry->value * LL[i];
        count++;
      }
    }
  }
  return LL;
}

template<class numeric> homopoly<numeric> Simp(monomial t, homopoly<numeric> f, list<list<homopoly<numeric> > > &F, list<list<homopoly<numeric> > > &Ftil) {
  if (t == monomial())
    return f;
  list<monomial> M;
  homopoly<numeric> p;
  monomial m;
  unsigned long d, i, j, k, l = F.length(), ll, n, s;
  bool degcomp = true;
  M.append(t);
  if (TermOrd != DegLex && TermOrd != DegRevLex) {
    degcomp = false;
    for (d = t.deg() - 1; d > 0; d--) {
      n = binomial(IndetList.entries + d - 1, d);
      for (i = 1; i <= n; i++) {
        m = indets.GetHomoMono(d, i);
        if (m.devides(t))
          M.append(m);
      }
    }
    M.sort('d');
  }
  else
    d = t.deg();
  while (d > 0 || !degcomp) {
    s = M.length();
    for (j = 1; j <= s; j++) {
      p = M[j] * f;
      for (k = 1; k <= l; k++) {
        ll = F[k].length();
        if (ll > 0 && F[k][1].deg() == p.deg())
          for (i = 1; i <= ll; i++)
            if (F[k][i] == p) {
              ll = Ftil[k].length();
              for (i = 1; i <= ll; i++)
                if (Ftil[k][i].LT() == p.LT())
                  if (M[j] == t)
                    return Ftil[k][i];
                  else
                    return Simp(t / M[j], Ftil[k][i], F, Ftil);
              cerr << "Severe error in function Simp! Element of F~+ not found!!!" << endl;
            }
      }
    }
    if (!degcomp)
      break;
    d--;
    if (d > 0) {
      M.clear();
      n = binomial(IndetList.entries + d - 1, d);
      for (i = 1; i <= n; i++) {
        m = indets.GetHomoMono(d, i);
        if (m.devides(t))
          M.append(m);
      }
    }
  }
  return t * f / f.LC();
}

template<class numeric> homopoly<numeric> Simp(monomial t, homopoly<numeric> f, list<list<homopoly<numeric> > > &F, list<list<homopoly<numeric> > > &Ftil, unsigned long &count) {
  count++;
  if (t == monomial())
    return f;
  list<monomial> M;
  homopoly<numeric> p;
  monomial m;
  unsigned long d, i, j, k, l = F.length(), ll, n, s;
  bool degcomp = true;
  M.append(t);
  if (TermOrd != DegLex && TermOrd != DegRevLex) {
    degcomp = false;
    for (d = t.deg() - 1; d > 0; d--) {
      n = binomial(IndetList.entries + d - 1, d);
      for (i = 1; i <= n; i++) {
        m = indets.GetHomoMono(d, i);
        if (m.devides(t))
          M.append(m);
      }
    }
    M.sort('d');
  }
  else
    d = t.deg();
  while (d > 0 || !degcomp) {
    s = M.length();
    for (j = 1; j <= s; j++) {
      p = M[j] * f;
      for (k = 1; k <= l; k++) {
        ll = F[k].length();
        if (ll > 0 && F[k][1].deg() == p.deg())
          for (i = 1; i <= ll; i++)
            if (F[k][i] == p) {
              ll = Ftil[k].length();
              for (i = 1; i <= ll; i++)
                if (Ftil[k][i].LT() == p.LT())
                  if (M[j] == t)
                    return Ftil[k][i];
                  else
                    return Simp(t / M[j], Ftil[k][i], F, Ftil, count);
              cerr << "Severe error in function Simp! Element of F~+ not found!!!" << endl;
            }
      }
    }
    if (!degcomp)
      break;
    d--;
    if (d > 0) {
      M.clear();
      n = binomial(IndetList.entries + d - 1, d);
      for (i = 1; i <= n; i++) {
        m = indets.GetHomoMono(d, i);
        if (m.devides(t))
          M.append(m);
      }
    }
  }
  return t * f / f.LC();
}


// sort polynomials

template<class polynomial> void PolySort(list<polynomial> &F, const unsigned long &start) {

  unsigned long i, j;
  polynomial zeropoly;
  for (i = F.length(); i > 0; i--)
    if (F[i] == zeropoly)
      F.remove(i);

  unsigned long k, l = F.length(), pos;
  monomial lead;
  list<polynomial> Fold = F;
  for (i = start; i <= l; i++) {
    k = Fold.length();
    lead = Fold[start].LT();
    pos = start;
    for (j = start + 1; j <= k; j++)
      if (Fold[j].LT() > lead) {
        lead = Fold[j].LT();
        pos = j;
      }
    F[i] = Fold[pos];
    Fold.remove(pos);
  }
}

#endif

