#ifndef poly_h
#define poly_h
#include <iostream>
#include <string>
#include "error.h"
#include "mathops.h"
#include "z.h"
#include "contain.t"
#include "sparse.t"

extern bool debugpoly;

/******************/
/* indeterminates */
/******************/

// list of all existing indeterminates:

class indet_list {
  friend class indet    ;
  friend class Indet_Ops;

  public:
    unsigned long   entries;
    char          * name   ;
    unsigned long * index  ;
    indet*        * point  ;

    // standard constructor:
    indet_list():
      entries(                  0 ) ,
      name   (new char         [0]) ,
      index  (new unsigned long[0]) ,
      point  (new indet*       [0]) {
    }

    // destructor:
    ~indet_list() {
      delete[] name ;
      delete[] index;
      delete[] point;
    }

    // methods:
    unsigned long search  (const char&                      );
    unsigned long search  (const char&, const unsigned long&);
    unsigned long searchex(const char&                      );

  private:
    // methods:
    void add   (      indet&);
    void clear (            );
    void remove(const indet&);
};

extern indet_list IndetList;

// class for indeterminates:

class indet {
  friend class indet_list                  ;
  friend class Indet_Ops                   ;
  friend class pp                          ;
  friend class monomial                    ;
  friend monomial LogToMono(const list<Z>&);

  public:
    // constructors:
    // standard constructor:
    indet() {
    }
    // generel constructors:
    indet(const char &c):
      name (c    ) ,
      index(false) ,
      begin(1    ) ,
      end  (0    ) {
      IndetList.add(*this);
    }
    indet(const char &c, const unsigned long &m, const unsigned long &n):
      name (c   ) ,
      index(true) ,
      begin(m   ) ,
      end  (n   ) {
      IndetList.add(*this);
    }

    // destructor:
    ~indet() {
      IndetList.remove(*this);
    }

    // methods:
    void reindex() {
      index = false;
      IndetList.add(*this);
    }
    void reindex(const unsigned long &m, const unsigned long &n) {
      index = true;
      begin = m   ;
      end   = n   ;
      IndetList.add(*this);
    }

  private:
    char name;
    bool index;
    unsigned long begin, end;
};

// class for indeterminate handling:

struct Indet_Listing {
  indet          ind ;
  Indet_Listing* next;

  // constructor:
  Indet_Listing() {next = NULL;}
};

class Indet_Ops {

  public:
    // standard constructor:
    Indet_Ops() {
      first = NULL;
      last  = NULL;
    }

    // destructor:
    ~Indet_Ops();

    // methods:
    void          add         (const char         &                                            );
    void          add         (const char         &, const unsigned long&                      );
    void          add         (const char         &, const unsigned long&, const unsigned long&);
    void          CalcMonoLine(const unsigned long&                                            );
    monomial      GetHomoMono (const unsigned long&, const unsigned long&                      );
    unsigned long GetHomoPos  (const monomial     &                                            );
    string        GetList     (                                                                );
    void          modify      (const char         &                                            );
    void          modify      (const char         &, const unsigned long&                      );
    void          modify      (const char         &, const unsigned long&, const unsigned long&);
    void          remove      (const char         &                                            );
    void          removeall   (                                                                );

  private:
    Indet_Listing *first, *last;
};

extern Indet_Ops               indets   ;
extern list<list<vector<Z> > > MonoTable;


/******************/
/* power products */
/******************/

class pp {
  friend class monomial;
  friend bool  operator==(const pp&, const pp&);

  public:
    string Indet   ;
    Z      Exponent;

    // constructors:
    // standard constructor:
    pp() {};
    // generel constructor:
    pp(const string&, const Z &a = 1);

    // internal operators:
    pp operator/=(const pp      &);
    pp operator/=(const monomial&);
    pp operator^=(const Z       &);

    // method:
    Z deg() const;

  private:
    indet*        ind  ;
    unsigned long index;
};

// external operators:

pp       operator/ (const pp&, const pp      &);
pp       operator/ (const pp&, const monomial&);
pp       operator^ (const pp&, const Z       &);
bool     operator==(const pp&, const pp      &);
bool     operator!=(const pp&, const pp      &);
ostream& operator<<(ostream &, const pp      &);


/*************/
/* monomials */ // here monomials are terms without coefficients
/*************/

class monomial {
  friend class    pp                                          ;
  friend bool     operator==(const monomial&, const monomial&);
  friend ostream& operator<<(      ostream &, const monomial&);

  public:
    unsigned long ppcount;
    list<pp>      pps    ;

    // constructors:

    // standard constructor:
    monomial():
      ppcount(0) {
    }

    // generel constructors:
    monomial(const string&);
    monomial(const pp    &);

    // internal operators:
    monomial operator*=(const pp      &);
    monomial operator*=(const monomial&);
    monomial operator/=(const pp      &);
    monomial operator/=(const monomial&);
    monomial operator^=(const Z       &);
    monomial operator =(const pp      &);
    monomial operator =(const monomial&);

    // methods:
    Z       deg    (               ) const;
    bool    devides(const monomial&) const;
    list<Z> log    (               ) const;
};

// external operators:

monomial operator* (const pp      &, const pp      &);
monomial operator* (const pp      &, const monomial&);
monomial operator* (const monomial&, const pp      &);
monomial operator* (const monomial&, const monomial&);
monomial operator/ (const monomial&, const pp      &);
monomial operator/ (const monomial&, const monomial&);
monomial operator^ (const monomial&, const Z       &);
bool     operator==(const monomial&, const monomial&);
bool     operator!=(const monomial&, const monomial&);
bool     operator< (const monomial&, const monomial&);
bool     operator> (const monomial&, const monomial&);
bool     operator<=(const monomial&, const monomial&);
bool     operator>=(const monomial&, const monomial&);
ostream& operator<<(      ostream &, const monomial&);

// functions:

monomial gcd      (const monomial&, const monomial&);
monomial lcm      (const monomial&, const monomial&);
monomial LogToMono(const list<Z> &                 );


/*********/
/* terms */ // here terms are monomials with coefficient
/*********/

template<class numeric> class term {
  friend class poly    <numeric>                                             ;
  friend class homopoly<numeric>                                             ;
  friend term<numeric> operator- (const term<numeric>&                      );
  friend bool          operator==(const term<numeric>&, const term<numeric>&);
  friend bool          operator< (const term<numeric>&, const term<numeric>&);
  friend ostream&      operator<<(      ostream      &, const term<numeric>&);
  friend ostream&      operator<<(      ostream      &, const poly<numeric>&);

  public:
    // constructors:

    // standard constructor:

    term():
      coeff(0) ,
      mono ( ) {
    }

    // generel constructors:

    term(const pp &x):
      coeff(1) ,
      mono (x) {
    }

    term(const numeric &a, const pp &x):
      coeff(a) ,
      mono (x) {
    }

    term(const monomial &m):
      coeff(1) ,
      mono (m) {
    }

    term(const numeric &a, const monomial &m = monomial()):
      coeff(a) ,
      mono (m) {
    }

    // internal operators:

    // addition and substraction operators:

    term operator+=(const term &t) {
      if (t.coeff == 0)
        return *this;
      if (coeff == 0) {
        coeff = t.coeff;
        mono  = t.mono ;
        return *this;
      }
      if (mono != t.mono)
        Error(TermNoAdd, "term+=term");
      coeff += t.coeff;
      return *this;
    }

    term operator-=(const term &t) {
      return *this += -t;
    }

    // multiplication and division operators:

    term operator*=(const numeric &a) {
      coeff *= a;
      return *this;
    }

    term operator*=(const pp &x) {
      mono *= x;
      return *this;
    }

    term operator*=(const monomial &m) {
      mono *= m;
      return *this;
    }

    term operator*=(const term &t) {
      coeff *= t.coeff;
      mono  *= t.mono ;
      return *this;
    }

    term operator/=(const numeric &a) {
      coeff /= a;
      return *this;
    }

    term operator/=(const pp &x) {
      unsigned long pos = 0;
      for (unsigned long i = 1; i <= mono.ppcount; i++)
        if (mono.pps[i].Indet == x.Indet) {
          pos = i;
          break;
        }
      if (pos == 0 || mono.pps[pos].Exponent < x.Exponent) {
        mono.pps.clear();
        mono.ppcount = 0;
        coeff = 0;
      }
      else
        if (mono.pps[pos].Exponent == x.Exponent) {
          mono.pps.remove(pos);
          mono.ppcount--;
        }
        else
          mono.pps[pos].Exponent -= x.Exponent;
      return *this;
    }

    term operator/=(const monomial &m) {
      for (unsigned i = 1; i <= m.ppcount; i++)
        *this /= m.pps[i];
      return *this;
    }

    term operator/=(const term &t) {
      *this /= t.mono;
      return *this /= t.coeff;
    }

    term operator%=(const numeric &a) {
      coeff %= a;
      return *this;
    }

    term operator^=(const Z &a) {
      coeff ^= a;
      mono  ^= a;
      return *this;
    }

    // assignment operators:

    term operator=(const pp &x) {
      coeff = 1;
      mono  = x;
      return *this;
    }

    term operator=(const monomial &m) {
      coeff = 1;
      mono  = m;
      return *this;
    }

    term operator=(const term &t) {
      if (*this != t) {
        coeff = t.coeff;
        mono  = t.mono ;
      }
      return *this;
    }

    // method:
    Z deg() const {
      return mono.deg();
    }

  private:
    numeric  coeff;
    monomial mono ;
};

// external operators:

// addition and substraction operators:

template<class numeric> term<numeric> operator+(const term<numeric> &t) {
  return t;
}

template<class numeric> term<numeric> operator-(const term<numeric> &t) {
  term<numeric> temp = t;
  temp.coeff = -temp.coeff;
  return temp;
}

// multiplication and division operators:

template<class numeric> term<numeric> operator*(const pp &x, const numeric &a) {
  return term<numeric>(a, x);
}

template<class numeric> term<numeric> operator*(const numeric &a, const pp &x) {
  return term<numeric>(a, x);
}
/**
template<class numeric> term<numeric> operator*(const monomial &m, const numeric &a) {
  return term<numeric>(a, m);
}

template<class numeric> term<numeric> operator*(const numeric &a, const monomial &m) {
  return term<numeric>(a, m);
}
**/
template<class numeric> term<numeric> operator*(const term<numeric> &t, const numeric &a) {
  term<numeric> temp = t;
  return temp *= a;
}

template<class numeric> term<numeric> operator*(const numeric &a, const term<numeric> &t) {
  term<numeric> temp = t;
  return temp *= a;
}

template<class numeric> term<numeric> operator*(const term<numeric> &t, const pp &x) {
  term<numeric> temp = t;
  return temp *= x;
}

template<class numeric> term<numeric> operator*(const pp &x, const term<numeric> &t) {
  term<numeric> temp = t;
  return temp *= x;
}

template<class numeric> term<numeric> operator*(const term<numeric> &t, const monomial &m) {
  term<numeric> temp = t;
  return temp *= m;
}

template<class numeric> term<numeric> operator*(const monomial &m, const term<numeric> &t) {
  term<numeric> temp = t;
  return temp *= m;
}

template<class numeric> term<numeric> operator*(const term<numeric> &s, const term<numeric> &t) {
  term<numeric> temp = s;
  return temp *= t;
}

template<class numeric> term<numeric> operator/(const term<numeric> &t, const numeric &a) {
  term<numeric> temp = t;
  return temp /= a;
}

template<class numeric> term<numeric> operator/(const term<numeric> &t, const pp &x) {
  term<numeric> temp = t;
  return temp /= x;
}

template<class numeric> term<numeric> operator/(const term<numeric> &t, const monomial &m) {
  term<numeric> temp = t;
  return temp /= m;
}

template<class numeric> term<numeric> operator/(const term<numeric> &s, const term<numeric> &t) {
  term<numeric> temp = s;
  return temp /= t;
}

template<class numeric> term<numeric> operator%(const term<numeric> &t, const numeric &a) {
  term<numeric> temp = t;
  return temp %= a;
}

template<class numeric> term<numeric> operator^(const term<numeric> &t, const Z &a) {
  term<numeric> temp = t;
  return temp ^= a;
}

// comparison operators:

template<class numeric> bool operator==(const term<numeric> &s, const term<numeric> &t) {
  if (s.coeff == t.coeff && s.mono == t.mono)
    return true;
  return false;
}

template<class numeric> bool operator!=(const term<numeric> &s, const term<numeric> &t) {
  return !(s == t);
}

template<class numeric> bool operator<(const term<numeric> &s, const term<numeric> &t) {
  if (s.mono == t.mono)
    return s.coeff < t.coeff;
  return s.mono < t.mono;
}

template<class numeric> bool operator>(const term<numeric> &s, const term<numeric> &t) {
  return t < s;
}

template<class numeric> bool operator<=(const term<numeric> &s, const term<numeric> &t) {
  return s == t || s < t;
}

template<class numeric> bool operator>=(const term<numeric> &s, const term<numeric> &t) {
  return s == t || t < s;
}

// output operator:

template<class numeric> ostream& operator<<(ostream &out, const term<numeric> &t) {
  if (t.mono.ppcount == 0)
    out << t.coeff;
  else  {
    if (t.coeff == 0)
      out << 0;
    else {
      if (t.coeff != 1)
        if (Field == RAT && t.coeff == -1)
          out << '-';
        else
          out << t.coeff;
      out << t.mono;
    }
  }
  return out;
}


/***************/
/* polynomials */
/***************/

template<class numeric> class poly {
  friend class homopoly<numeric>                                             ;
  friend poly<numeric> operator- (const poly<numeric>&                      );
  friend bool          operator==(const poly<numeric>&, const poly<numeric>&);
  friend ostream&      operator<<(      ostream      &, const poly<numeric>&);

  public:
    // constructors:

    // standard constructor:

    poly():
      termcount(0   ) ,
      leader   (NULL) {
    }

    // generel constructors:

    poly(const numeric &a):
      termcount(a != numeric(0)) ,
      leader   (NULL           ) {
      if (termcount)
        terms.append(term<numeric>(a));
    }

    poly(const pp &x):
      termcount(1   ) ,
      leader   (NULL) {
      terms.append(term<numeric>(x));
    }

    poly(const monomial &m):
      termcount(1   ) ,
      leader   (NULL) {
      terms.append(term<numeric>(m));
    }

    poly(const term<numeric> &t):
      termcount(t != term<numeric>(0)) ,
      leader   (NULL                 ) {
      if (termcount)
        terms.append(t);
    }

    // copy constructor:

    poly(const poly<numeric> &p):
      termcount(p.termcount) ,
      terms    (p.terms    ) ,
      leader   (NULL       ) {
    }

    // internal operators:

    // addition and substraction operators:

    poly operator+=(const poly &p) {
      unsigned long pos;
      leader = NULL;
      for (unsigned long i = 1; i <= p.termcount; i++) {
        for (pos = 1; pos <= termcount; pos++)
          if (terms[pos].mono == p.terms[i].mono)
            break;
        if (pos > termcount) {
          terms.append(p.terms[i]);
          termcount++;
        }
        else
          if (terms[pos].coeff == -p.terms[i].coeff) {
            terms.remove(pos);
            termcount--;
          }
          else
            terms[pos].coeff += p.terms[i].coeff;
      }
      return *this;
    }

    poly operator-=(const poly &p) {
      return *this += -p;
    }

    // multiplication and division operators:

    poly operator*=(poly p) {
      poly temp = *this;
      terms.clear();
      termcount = 0;
      leader = NULL;
      for (unsigned long i = 1; i <= temp.termcount; i++)
        for (unsigned long j = 1; j <= p.termcount; j++)
          *this += temp.terms[i] * p.terms[j];
      return *this;
    }

    poly operator/=(const numeric &a) {
      for (unsigned long i = 1; i <= termcount; i++)
        terms[i].coeff /= a;
      return *this;
    }

    poly operator/=(const pp &x) {
      for (unsigned long i = termcount; i > 0; i--) {
        terms[i] /= x;
        if (terms[i].coeff == 0) {
          terms.remove(i);
          termcount--;
        }
      }
      return *this;
    }

    poly operator/=(const monomial &m) {
      for (unsigned long i = termcount; i > 0; i--) {
        terms[i] /= m;
        if (terms[i].coeff == 0) {
          terms.remove(i);
          termcount--;
        }
      }
      return *this;
    }

    poly operator/=(const term<numeric> &t) {
      *this /= t.mono;
      return *this /= t.coeff;
    }

    poly operator/=(const poly &p) {
      *this = Div(p)[1];
      leader = NULL;
      return *this;
    }

    poly operator%=(const numeric &a) {
      poly temp = *this / a;
      temp *= a;
      return *this -= temp;
    }

    poly operator%=(const pp &x) {
      poly temp = *this / x;
      temp *= x;
      return *this -= temp;
    }

    poly operator%=(const monomial &m) {
      poly temp = *this / m;
      temp *= m;
      return *this -= temp;
    }

    poly operator%=(const term<numeric> &t) {
      poly temp = *this / t;
      temp *= t;
      return *this -= temp;
    }

    poly operator%=(const poly &p) {
      *this = NR(p);
      leader = NULL;
      return *this;
    }

    poly operator^=(Z a) {
      if (a < 0)
        Error(ExpLessZero, "poly^=Z");
      poly mult = *this, prod = numeric(1);
      for (;a != 0; a /= 2, mult *= mult)
        if (a % 2 == 1)
          prod = mult * prod;
      *this = prod;
      return prod;
    }

    // assignment operators:

    poly operator=(const numeric &a) {
      termcount = a != numeric(0);
      terms.clear();
      if (termcount) {
        terms.append(term<numeric>(a));
        leader = &terms[1];
      }
      else
        leader = NULL;
      return *this;
    }

    poly operator=(const pp &x) {
      termcount = 1;
      terms.clear();
      terms.append(term<numeric>(x));
      leader = &terms[1];
      return *this;
    }

    poly operator=(const monomial &m) {
      termcount = 1;
      terms.clear();
      terms.append(term<numeric>(m));
      leader = &terms[1];
      return *this;
    }

    poly operator=(const term<numeric> &t) {
      termcount = t != term<numeric>(0);
      terms.clear();
      if (termcount) {
        terms.append(t);
        leader = &terms[1];
      }
      else
        leader = NULL;
      return *this;
    }

    poly operator=(const poly &p) {
      if (*this != p) {
        termcount = p.termcount;
        terms     = p.terms    ;
        leader    = NULL       ;
      }
      return *this;
    }


    // methods:

    // degree
    Z deg() const {
      Z degree = 0;
      for (unsigned long i = 1; i <= termcount; i++)
        degree = max(degree, terms[i].mono.deg());
      return degree;
    }

    // homogeneous?
    bool ishomo() const {
      if (termcount > 1) {
        Z d = terms[1].mono.deg();
        for (unsigned long i = 2; i <= termcount; i++)
          if (terms[i].mono.deg() != d)
            return false;
      }
      return true;
    }

    // leading coefficient
    numeric LC() {
      FindLead();
      if (leader == NULL)
        return numeric(0);
      else
        return (*leader).coeff;
    }

    // leading monomial (which is a "term" here)
    term<numeric> LM() {
      FindLead();
      if (leader == NULL)
        return term<numeric>(0);
      else
        return *leader;
    }

    // leading term (which is a "monomial" here)
    monomial LT() {
      FindLead();
      if (leader == NULL)
        return monomial();
      else
        return (*leader).mono;
    }

    // normal reduction
    poly NR(const poly &p) const {
      return NR(list<poly>(1, p));
    }

    poly StatNR(const poly &p, unsigned long &count) const {
      return StatNR(list<poly>(1, p), count);
    }

    poly NR(list<poly> g) const {
      unsigned long i, s = g.length();
      poly v = *this, q, diff, zeropoly;
      term<numeric> termdiv;
      monomial LTv;

      do {
        do {
          LTv = v.LT();
          for (i = 1; i <= s; i++) {
            if (g[i].LT().devides(LTv)) {
              termdiv = v.LM() / g[i].LM();
              diff = termdiv * g[i];
              v -= diff;
              break;
            }
          }
        } while (i <= s && v != zeropoly);
        q += v.LM();
        v -= v.LM();
      } while (v != zeropoly);
      return q;
    }

    poly StatNR(list<poly> g, unsigned long &count) const {
      unsigned long i, s = g.length();
      poly v = *this, q, zeropoly;
      term<numeric> termdiv;
      monomial LTv;

      do {
        do {
          LTv = v.LT();
          for (i = 1; i <= s; i++) {
            if (g[i].LT().devides(LTv)) {
              count++;
              termdiv = v.LM() / g[i].LM();
              v -= termdiv * g[i];
              break;
            }
          }
        } while (i <= s && v != zeropoly);
        q += v.LM();
        v -= v.LM();
      } while (v != zeropoly);
      return q;
    }

    // support
    list<monomial> supp() const {
      list<monomial> support(termcount, monomial("1"));
      for (unsigned long i = 1; i <= termcount; i++)
        support[i] = terms[i].mono;
      return support;
    }

  private:
    typedef term<numeric> numericterm;
    unsigned long     termcount;
    list<numericterm> terms    ;
    numericterm*      leader   ;

    // methods:

    // division algorithm
    list<poly> Div(const poly &p) const {
      return Div(list<poly>(1, p));
    }

    list<poly> Div(list<poly> g) const {
      unsigned long i, q = g.length() + 1;
      poly v = *this, zeropoly;
      term<numeric> termdiv;
      monomial LTv;
      list<poly> p(q, zeropoly);

      do {
        do {
          LTv = v.LT();
          for (i = 1; i < q; i++) {
            if (g[i].LT().devides(LTv)) {
              termdiv = v.LM() / g[i].LM();
              p[i] += termdiv       ;
              v    -= termdiv * g[i];
              break;
            }
          }
        } while (i < q);
        p[q] += v.LM();
        v    -= v.LM();
      } while (v != zeropoly);
      return p;
    }

    // lead term scan
    void FindLead() {
      if (leader == NULL)
        if (termcount > 0) {
          leader = &terms[1];
          for (unsigned long i = 2; i <= termcount; i++)
            if (*leader < terms[i])
              leader = &terms[i];
        }
    }
};

// external operators:

// addition and substraction operators:

template<class numeric> poly<numeric> operator+(const poly<numeric> &p) {
  return p;
}

template<class numeric> poly<numeric> operator-(const poly<numeric> &p) {
  poly<numeric> temp = p;
  for (unsigned long i = 1; i <= temp.termcount; i++)
    temp.terms[i] = -temp.terms[i];
  return temp;
}

template<class numeric> poly<numeric> operator+(const term<numeric> &t, const numeric &a) {
  poly<numeric> temp = t;
  return temp += a;
}

template<class numeric> poly<numeric> operator+(const numeric &a, const term<numeric> &t) {
  poly<numeric> temp = a;
  return temp += t;
}

template<class numeric> poly<numeric> operator+(const term<numeric> &t, const pp &x) {
  poly<numeric> temp = t;
  return temp += x;
}

template<class numeric> poly<numeric> operator+(const pp &x, const term<numeric> &t) {
  poly<numeric> temp = x;
  return temp += t;
}

template<class numeric> poly<numeric> operator+(const term<numeric> &t, const monomial &m) {
  poly<numeric> temp = t;
  return temp += m;
}

template<class numeric> poly<numeric> operator+(const monomial &m, const term<numeric> &t) {
  poly<numeric> temp = m;
  return temp += t;
}

template<class numeric> poly<numeric> operator+(const term<numeric> &s, const term<numeric> &t) {
  poly<numeric> temp = s;
  return temp += t;
}

template<class numeric> poly<numeric> operator+(const poly<numeric> &p, const numeric &a) {
  poly<numeric> temp = p;
  return temp += a;
}

template<class numeric> poly<numeric> operator+(const numeric &a, const poly<numeric> &p) {
  poly<numeric> temp = a;
  return temp += p;
}

template<class numeric> poly<numeric> operator+(const poly<numeric> &p, const pp &x) {
  poly<numeric> temp = p;
  return temp += x;
}

template<class numeric> poly<numeric> operator+(const pp &x, const poly<numeric> &p) {
  poly<numeric> temp = x;
  return temp += p;
}

template<class numeric> poly<numeric> operator+(const poly<numeric> &p, const monomial &m) {
  poly<numeric> temp = p;
  return temp += m;
}

template<class numeric> poly<numeric> operator+(const monomial &m, const poly<numeric> &p) {
  poly<numeric> temp = m;
  return temp += p;
}

template<class numeric> poly<numeric> operator+(const poly<numeric> &p, const term<numeric> &t) {
  poly<numeric> temp = p;
  return temp += t;
}

template<class numeric> poly<numeric> operator+(const term<numeric> &t, const poly<numeric> &p) {
  poly<numeric> temp = t;
  return temp += p;
}

template<class numeric> poly<numeric> operator+(const poly<numeric> &p, const poly<numeric> &q) {
  poly<numeric> temp = p;
  return temp += q;
}

template<class numeric> poly<numeric> operator-(const term<numeric> &t, const numeric &a) {
  poly<numeric> temp = t;
  return temp -= a;
}

template<class numeric> poly<numeric> operator-(const numeric &a, const term<numeric> &t) {
  poly<numeric> temp = a;
  return temp += -t;
}

template<class numeric> poly<numeric> operator-(const term<numeric> &t, const pp &x) {
  poly<numeric> temp = t;
  return temp -= x;
}

template<class numeric> poly<numeric> operator-(const pp &x, const term<numeric> &t) {
  poly<numeric> temp = x;
  return temp += -t;
}

template<class numeric> poly<numeric> operator-(const term<numeric> &t, const monomial &m) {
  poly<numeric> temp = t;
  return temp -= m;
}

template<class numeric> poly<numeric> operator-(const monomial &m, const term<numeric> &t) {
  poly<numeric> temp = m;
  return temp += -t;
}

template<class numeric> poly<numeric> operator-(const term<numeric> &s, const term<numeric> &t) {
  poly<numeric> temp = s;
  return temp += -t;
}

template<class numeric> poly<numeric> operator-(const poly<numeric> &p, const numeric &a) {
  poly<numeric> temp = p;
  return temp -= a;
}

template<class numeric> poly<numeric> operator-(const numeric &a, const poly<numeric> &p) {
  poly<numeric> temp = a;
  return temp += -p;
}

template<class numeric> poly<numeric> operator-(const poly<numeric> &p, const pp &x) {
  poly<numeric> temp = p;
  return temp -= x;
}

template<class numeric> poly<numeric> operator-(const pp &x, const poly<numeric> &p) {
  poly<numeric> temp = x;
  return temp += -p;
}

template<class numeric> poly<numeric> operator-(const poly<numeric> &p, const monomial &m) {
  poly<numeric> temp = p;
  return temp -= m;
}

template<class numeric> poly<numeric> operator-(const monomial &m, const poly<numeric> &p) {
  poly<numeric> temp = m;
  return temp += -p;
}

template<class numeric> poly<numeric> operator-(const poly<numeric> &p, const term<numeric> &t) {
  poly<numeric> temp = p;
  return temp += -t;
}

template<class numeric> poly<numeric> operator-(const term<numeric> &t, const poly<numeric> &p) {
  poly<numeric> temp = t;
  return temp += -p;
}

template<class numeric> poly<numeric> operator-(const poly<numeric> &p, const poly<numeric> &q) {
  poly<numeric> temp = p;
  return temp += -q;
}

// multiplication and division operators:

template<class numeric> poly<numeric> operator*(const poly<numeric> &p, const numeric &a) {
  poly<numeric> temp = p;
  return temp *= a;
}

template<class numeric> poly<numeric> operator*(const numeric &a, const poly<numeric> &p) {
  poly<numeric> temp = p;
  return temp *= a;
}

template<class numeric> poly<numeric> operator*(const poly<numeric> &p, const pp &x) {
  poly<numeric> temp = p;
  return temp *= x;
}

template<class numeric> poly<numeric> operator*(const pp &x, const poly<numeric> &p) {
  poly<numeric> temp = p;
  return temp *= x;
}

template<class numeric> poly<numeric> operator*(const poly<numeric> &p, const monomial &m) {
  poly<numeric> temp = p;
  return temp *= m;
}

template<class numeric> poly<numeric> operator*(const monomial &m, const poly<numeric> &p) {
  poly<numeric> temp = p;
  return temp *= m;
}

template<class numeric> poly<numeric> operator*(const poly<numeric> &p, const term<numeric> &t) {
  poly<numeric> temp = p;
  return temp *= t;
}

template<class numeric> poly<numeric> operator*(const term<numeric> &t, const poly<numeric> &p) {
  poly<numeric> temp = p;
  return temp *= t;
}

template<class numeric> poly<numeric> operator*(const poly<numeric> &p, const poly<numeric> &q) {
  poly<numeric> temp = p;
  return temp *= q;
}

template<class numeric> poly<numeric> operator/(const poly<numeric> &p, const numeric &a) {
  poly<numeric> temp = p;
  return temp /= a;
}

template<class numeric> poly<numeric> operator/(const poly<numeric> &p, const pp &x) {
  poly<numeric> temp = p;
  return temp /= x;
}

template<class numeric> poly<numeric> operator/(const poly<numeric> &p, const monomial &m) {
  poly<numeric> temp = p;
  return temp /= m;
}

template<class numeric> poly<numeric> operator/(const poly<numeric> &p, const term<numeric> &t) {
  poly<numeric> temp = p;
  return temp /= t;
}

template<class numeric> poly<numeric> operator/(const poly<numeric> &p, const poly<numeric> &q) {
  poly<numeric> temp = p;
  return temp /= q;
}

template<class numeric> poly<numeric> operator%(const poly<numeric> &p, const numeric &a) {
  poly<numeric> temp = p;
  return temp %= a;
}

template<class numeric> poly<numeric> operator%(const poly<numeric> &p, const pp &x) {
  poly<numeric> temp = p;
  return temp %= x;
}

template<class numeric> poly<numeric> operator%(const poly<numeric> &p, const monomial &m) {
  poly<numeric> temp = p;
  return temp %= m;
}

template<class numeric> poly<numeric> operator%(const poly<numeric> &p, const term<numeric> &t) {
  poly<numeric> temp = p;
  return temp %= t;
}

template<class numeric> poly<numeric> operator%(const poly<numeric> &p, const poly<numeric> &q) {
  poly<numeric> temp = p;
  return temp %= q;
}

template<class numeric> poly<numeric> operator^(const poly<numeric> &p, const Z &a) {
  poly<numeric> temp = p;
  return temp ^= a;
}

// comparison operators:

template<class numeric> bool operator==(const poly<numeric> &p, const poly<numeric> &q) {
  if (p.termcount != q.termcount)
    return false;
  for (unsigned long i = 1; i <= p.termcount; i++)
    if (!q.terms.seek(p.terms[i]))
      return false;
  return true;
}

template<class numeric> bool operator!=(const poly<numeric> &p, const poly<numeric> &q) {
  return !(p == q);
}

template<class numeric> bool operator<(poly<numeric> &p, poly<numeric> &q) {
  if (p == poly<numeric>(0) && q == poly<numeric>(0))
    return false;
  if (p.LM() == q.LM()) {
    poly<numeric> pp(p), qq(q);
    pp -= pp.LM();
    qq -= qq.LM();
    return pp < qq;
  }
  return p.LM() < q.LM();
}

template<class numeric> bool operator>(poly<numeric> &p, poly<numeric> &q) {
  return q < p;
}

template<class numeric> bool operator<=(poly<numeric> &p, poly<numeric> &q) {
  return p == q || p < q;
}

template<class numeric> bool operator>=(poly<numeric> &p, poly<numeric> &q) {
  return p == q || q < p;
}

// output operator:

template<class numeric> ostream& operator<<(ostream &out, const poly<numeric> &p) {
  if (p.termcount == 0)
    out << 0;
  else
    for (unsigned long i = 1; i <= p.termcount; i++) {
      if (p.terms[i].coeff == 0)
        break;
      if (p.terms[i].coeff < 0) {
        if (i > 1)
          out << ' ';
        out << "- " << -p.terms[i];
      }
      else {
        if (i > 1)
          out << " + ";
        out << p.terms[i];
      }
    }
  return out;
}

// functions:

template<class numeric> poly<numeric> normal(poly<numeric> &p) {
  numeric leadcoeff = p.LC();
  poly<numeric> temp = p / leadcoeff;
  if (leadcoeff * temp == p)
    return temp;
  return p;
}


/**************************/
/* homogenous polynomials */
/**************************/

template<class numeric> class homopoly {
  friend class poly<numeric>;
  friend homopoly<numeric>        operator-            (const homopoly    <numeric>&                                                                      );
  friend bool                     operator==           (const homopoly    <numeric>&, const homopoly<numeric> &                                           );
  friend ostream&                 operator<<           (      ostream              &, const homopoly<numeric> &                                           );
  friend list<homopoly<numeric> > GBBuchbergerGauss    (const homoideal   <numeric>&                                                                      );
  friend list<homopoly<numeric> > interreduce          (list<homopoly<numeric> >                                                                          );
  friend list<homopoly<numeric> > interreduce          (list<homopoly<numeric> >    ,       unsigned long     &                                           );
  friend sparsematrix<numeric>    MatrixUpdate         (const sparsematrix<numeric>&, const unsigned long     &, list<homopoly<numeric> >&                );
  friend void                     MR                   (list<homopoly<numeric> >   &, list<homopoly<numeric> >&, const unsigned long     &                );
  friend void                     MR                   (list<homopoly<numeric> >   &, list<homopoly<numeric> >&, const unsigned long     &, unsigned long&);
  friend list<homopoly<numeric> > StatGBBuchbergerGauss(const homoideal   <numeric>&                                                                      );

  typedef poly<numeric> numpoly;

  public:
    // constructors:

    // standard constructor:

    homopoly():
      degree  (                      0 ) ,
      fields  (                      1 ) ,
      coeffs  (sparsevector<numeric>(1)) ,
      leadmono(term        <numeric>(0)) {
    }

    // generel constructors:

    homopoly(const numeric &a):
      degree  (                      0    ) ,
      fields  (                      1    ) ,
      coeffs  (sparsevector<numeric>(1, a)) ,
      leadmono(term        <numeric>(   a)) {
    }

    homopoly(const pp &x):
      degree(x.deg()) {
      fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
      coeffs = sparsevector<numeric>(fields);
      coeffs.first = new sparseentry<numeric>(indets.GetHomoPos(x), 1);
      leadmono = x;
    }

    homopoly(const monomial &m):
      degree(m.deg()) {
      fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
      coeffs = sparsevector<numeric>(fields);
      coeffs.first = new sparseentry<numeric>(indets.GetHomoPos(m), 1);
      leadmono = m;
    }

    homopoly(const term<numeric> &t):
      degree(t.deg()) {
      fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
      coeffs = sparsevector<numeric>(fields);
      coeffs.first = new sparseentry<numeric>(indets.GetHomoPos(t.mono), t.coeff);
      leadmono = t;
    }

    homopoly(poly<numeric> p):
      degree(p.deg()) {
      fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
      coeffs = sparsevector<numeric>(fields);
      for (unsigned long i = 1; i <= p.termcount; i++) {
        if (p.terms[i].deg() != degree)
          Error(PolyNotHomo, "homopoly()");
        coeffs.insert(indets.GetHomoPos(p.terms[i].mono), p.terms[i].coeff);
      }
      leadmono = p.LM();
    }

    homopoly(const Z &d, const sparsevector<numeric> &v):
      degree  (d                                                                      ) ,
      fields  (v.length()                                                             ) ,
      coeffs  (v                                                                      ) ,
      leadmono(term<numeric>(v.first->value, indets.GetHomoMono(d, v.first->position))) {
    }

    // copy constructor

/*
    homopoly(const homopoly<numeric> &p):
      degree  (p.degree  ) ,
      fields  (p.fields  ) ,
      coeffs  (p.coeffs  ) ,
      leadmono(p.leadmono) {
if (debug)
cerr << "      homoconst" << endl;
    }
*/

    // internal operators:

    // addition and substraction operators:

    homopoly operator+=(const monomial &m) {
      if (degree == 0 && coeffs[1] == 0) {
        degree = m.deg();
        fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
        coeffs = sparsevector<numeric>(fields);
        coeffs.clear();
        coeffs.append(indets.GetHomoPos(m), 1);
        leadmono = m;
        return *this;
      }
      if (m.deg() != degree)
        Error(PolyDiffDeg, "homopoly+=monomial");
      if (-leadmono == m && coeffs.first->next == NULL) {
        degree   = 0;
        fields   = 1;
        coeffs   = sparsevector<numeric>(1);
        leadmono = term<numeric>();
        return *this;
      }
      coeffs.add(indets.GetHomoPos(m), 1);
      if (leadmono.mono == m) {
        leadmono += m;
        if (leadmono.coeff == 0)
          findLM();
      }
      else
        if (leadmono < m)
          leadmono = m;
      return *this;
    }

    homopoly operator+=(const term<numeric> &t) {
      if (t == term<numeric>(0))
        return *this;
      if (degree == 0 && coeffs[1] == 0) {
        degree = t.deg();
        fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
        coeffs = sparsevector<numeric>(fields);
        coeffs.clear();
        coeffs.append(indets.GetHomoPos(t.mono), t.coeff);
        leadmono = t;
        return *this;
      }
      if (t.deg() != degree)
        Error(PolyDiffDeg, "homopoly+=term");
      if (leadmono == -t && coeffs.first->next == NULL) {
        degree   = 0;
        fields   = 1;
        coeffs   = sparsevector<numeric>(1);
        leadmono = term<numeric>();
        return *this;
      }
      coeffs.add(indets.GetHomoPos(t.mono), t.coeff);
      if (leadmono.mono == t.mono) {
        leadmono += t;
        if (leadmono.coeff == 0)
          findLM();
      }
      else
        if (leadmono < t)
          leadmono = t;
      return *this;
    }

    homopoly operator+=(const homopoly &p) {
      if (p.degree == 0 && p.coeffs[1] == 0)
        return *this;
      if (degree == 0 && coeffs[1] == 0) {
        degree   = p.degree;
        fields   = p.fields;
        coeffs   = p.coeffs;
        leadmono = p.leadmono;
        return *this;
      }
      if (p.degree != degree)
        Error(PolyDiffDeg, "homopoly+=homopoly");
      if (coeffs == -p.coeffs) {
        degree   = 0;
        fields   = 1;
        coeffs   = sparsevector<numeric>(1);
        leadmono = term<numeric>();
        return *this;
      }
      coeffs += p.coeffs;
      if (leadmono.mono == p.leadmono.mono) {
        leadmono += p.leadmono;
        if (leadmono.coeff == 0)
          findLM();
      }
      else
        if (leadmono < p.leadmono)
          leadmono = p.leadmono;
      return *this;
    }

    homopoly operator-=(const monomial &m) {
      return *this += -term<numeric>(m);
    }

    homopoly operator-=(const term<numeric> &t) {
      return *this += -t;
    }

    homopoly operator-=(const homopoly &p) {
      return *this += -p;
    }

    // multiplication and division operators:

    homopoly operator*=(const numeric &a) {
      if (a == 0) {
        degree   =                       0 ;
        fields   =                       1 ;
        coeffs   = sparsevector<numeric>(1);
        leadmono = term        <numeric>(0);
      }
      else {
        coeffs   *= a;
        leadmono *= a;
      }
      return *this;
    }

    homopoly operator*=(const pp &x) {
      return *this *= monomial(x);
    }

    homopoly operator*=(const monomial &m) {
      if (*this == homopoly())
        return *this;
      unsigned long k = 1;
      Z d = m.deg();
      if (d != 0) {
        indets.CalcMonoLine(degree    );
        indets.CalcMonoLine(degree + d);
        list<vector<Z> >      *LThis   = &MonoTable[degree    ],
                              *LResult = &MonoTable[degree + d];
        vector<Z>             VMult    = m.log()               ,
                              VResult                          ;
        sparsevector<numeric> newcoeffs(LResult->length())     ;
        sparseentry <numeric> *entry = coeffs.first            ;
        while (entry) {
          VResult = (*LThis)[entry->position] + VMult;
          while ((*LResult)[k] != VResult)
            k++;
          newcoeffs.append(k, entry->value);
          k++;
          entry = entry->next;
        }
        degree   += d               ;
        fields    = LResult->length();
        coeffs    = newcoeffs       ;
        leadmono *= m               ;
      }
      return *this;
    }

    homopoly operator*=(const term<numeric> &t) {
      if (*this == homopoly())
        return *this;
      unsigned long k = 1;
      Z d = t.deg();
      if (d == 0) {
        *this *= t.coeff;
        return *this;
      }
      else {
        indets.CalcMonoLine(degree    );
        indets.CalcMonoLine(degree + d);
        list<vector<Z> >      LThis   = MonoTable[degree    ],
                              LResult = MonoTable[degree + d];
        vector<Z>             VMult   = t.mono.log()         ,
                              VResult                        ;
        sparsevector<numeric> newcoeffs(LResult.length())    ;
        sparseentry <numeric> *entry = coeffs.first          ;
        while (entry) {
          VResult = LThis[entry->position] + VMult;
          while (LResult[k] != VResult)
            k++;
          newcoeffs.append(k, entry->value * t.coeff);
          k++;
          entry = entry->next;
        }
        degree   += d               ;
        fields    = LResult.length();
        coeffs    = newcoeffs       ;
        leadmono *= t               ;
      }
      return *this;
    }


    homopoly operator*=(const homopoly &p) {
      if (*this == homopoly())
        return *this;
      unsigned long k;
      Z d = p.deg();
      if (d == 0)
        if (p == homopoly()) {
          degree   =                       0 ;
          fields   =                       1 ;
          coeffs   = sparsevector<numeric>(1);
          leadmono = term        <numeric>(0);
        }
        else {
          coeffs   *= p.coeffs[1];
          leadmono *= p.coeffs[1];
        }
      else {
        indets.CalcMonoLine(degree    );
        indets.CalcMonoLine(         d);
        indets.CalcMonoLine(degree + d);
        list<vector<Z> >      LThis   = MonoTable[degree    ],
                              LMult   = MonoTable[         d],
                              LResult = MonoTable[degree + d];
        vector<Z>             VResult                        ;
        sparsevector<numeric> newcoeffs(LResult.length())    ;
        sparseentry <numeric> *entry                         ,
                              *pentry = p.coeffs.first       ;
        while (pentry) {
          k     = pentry->position;
          entry = coeffs.first    ;
          while (entry) {
            VResult = LThis[entry->position] + LMult[pentry->position];
            while (LResult[k] != VResult)
              k++;
            newcoeffs.add(k, entry->value * pentry->value);
            k++;
            entry = entry->next;
          }
          pentry = pentry->next;
        }
        degree   += d               ;
        fields    = LResult.length();
        coeffs    = newcoeffs       ;
        leadmono *= p.leadmono      ;
      }
      return *this;
    }

    homopoly operator/=(const numeric &a) {
      coeffs   /= a;
      leadmono /= a;
      if (coeffs.isnull()) {
        degree = 0;
        fields = 1;
        coeffs = sparsevector<numeric>(1);
      }
      return *this;
    }

    homopoly operator/=(const pp &x) {
      return *this /= monomial(x);
    }

    homopoly operator/=(const monomial &m) {
      if (*this == homopoly())
        return *this;
      bool positive;
      unsigned long  i, k = 1, l, pos;
      Z d = m.deg();
      if (d > degree) {
        degree   = 0                       ;
        fields   = 1                       ;
        coeffs   = sparsevector<numeric>(1);
        leadmono = term        <numeric>( );
      }
      else if (d == degree) {
        i        = indets.GetHomoPos(m)         ;
        degree   = 0                                  ;
        fields   = 1                                  ;
        coeffs   = sparsevector<numeric>(1, coeffs[i]);
        leadmono = term        <numeric>(   coeffs[1]);
      }
      else if (d != 0) {
        indets.CalcMonoLine(degree    );
        indets.CalcMonoLine(degree - d);
        list<vector<Z> >      LThis   = MonoTable[degree    ],
                              LResult = MonoTable[degree - d];
        vector<Z>             VDiv    = m.log()              ,
                              VResult                        ;
        sparsevector<numeric> newcoeffs(LResult.length())    ;
        sparseentry <numeric> *entry = coeffs.first          ;
        l = VDiv.length();
        while (entry) {
          VResult = LThis[entry->position] - VDiv;
          positive = true;
          for (pos = 1; pos <= l; pos++)
            if (VResult[pos] < 0) {
              positive = false;
              break;
            }
          if (positive) {
            while (LResult[k] != VResult)
              k++;
            newcoeffs.append(k, entry->value);
            k++;
          }
          entry = entry->next;
        }
        degree   -= d               ;
        fields    = LResult.length();
        coeffs    = newcoeffs       ;
        leadmono /= m               ;
      }
      return *this;
    }

    homopoly operator/=(const term<numeric> &t) {
      if (*this == homopoly())
        return *this;
      bool positive;
      unsigned long i, k = 1, l, pos;
      Z d = t.deg();
      if (d > degree) {
        degree   = 0                       ;
        fields   = 1                       ;
        coeffs   = sparsevector<numeric>(1);
        leadmono = term        <numeric>( );
      }
      else if (d == degree) {
        i        = indets.GetHomoPos(t.mono)                    ;
        degree   = 0                                            ;
        fields   = 1                                            ;
        coeffs   = sparsevector<numeric>(1, coeffs[i] / t.coeff);
        leadmono = term        <numeric>(   coeffs[1]          );
      }
      else if (d == 0) {
        coeffs   /= t.coeff;
        leadmono /= t.coeff;
      }
      else {
        indets.CalcMonoLine(degree    );
        indets.CalcMonoLine(degree - d);
        list<vector<Z> >      LThis   = MonoTable[degree    ],
                              LResult = MonoTable[degree - d];
        vector<Z>             VDiv    = t.mono.log()         ,
                              VResult                        ;
        sparsevector<numeric> newcoeffs(LResult.length())    ;
        sparseentry <numeric> *entry = coeffs.first          ;
        l = VDiv.length();
        while (entry) {
          VResult = LThis[entry->position] - VDiv;
          positive = true;
          for (pos = 1; pos <= l; pos++)
            if (VResult[pos] < 0) {
              positive = false;
              break;
            }
          if (positive) {
            while (LResult[k] != VResult)
              k++;
            newcoeffs.append(k, coeffs[i] / t.coeff);
            k++;
          }
          entry = entry->next;
        }
        degree   -= d               ;
        fields    = LResult.length();
        coeffs    = newcoeffs       ;
        leadmono /= t               ;
      }
      return *this;
    }

    homopoly operator/=(const homopoly &p) {
      *this = Div(p)[1];
      findLM();
      return *this;
    }

    homopoly operator%=(const numeric &a) {
      homopoly temp = *this;
      temp /= a;
      temp *= a;
      return *this -= temp;
    }

    homopoly operator%=(const pp &x) {
      homopoly temp = *this;
      temp /= x;
      temp *= x;
      return *this -= temp;
    }

    homopoly operator%=(const monomial &m) {
      homopoly temp = *this;
      temp /= m;
      temp *= m;
      return *this -= temp;
    }

    homopoly operator%=(const term<numeric> &t) {
      homopoly temp = *this;
      temp /= t;
      temp *= t;
      return *this -= temp;
    }

    homopoly operator%=(const homopoly &p) {
      *this = NR(p);
      findLM();
      return *this;
    }

    homopoly operator^=(Z a) {
      poly<numeric> temp(*this);
      temp ^= a;
      return homopoly(temp);
    }

    // assignment operators:

    homopoly operator=(const numeric &a) {
      degree   =                       0    ;
      fields   =                       1    ;
      coeffs   = sparsevector<numeric>(1, a);
      leadmono = term        <numeric>(   a);
      return *this;
    }

    homopoly operator=(const pp &x) {
      degree = x.deg();
      fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
      coeffs = sparsevector<numeric>(fields);
      coeffs.append(indets.GetHomoPos(x), 1);
      leadmono = x;
      return *this;
    }

    homopoly operator=(const monomial &m) {
      degree = m.deg();
      fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
      coeffs = sparsevector<numeric>(fields);
      coeffs.append(indets.GetHomoPos(m), 1);
      leadmono = m;
      return *this;
    }

    homopoly operator=(const term<numeric> &t) {
      degree = t.deg();
      fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
      coeffs = sparsevector<numeric>(fields);
      coeffs.append(indets.GetHomoPos(t.mono), t.coeff);
      leadmono = t;
      return *this;
    }

    homopoly operator=(poly<numeric> &p) {
      degree = p.deg();
      fields = binomial(IndetList.entries + (unsigned long)degree - 1, IndetList.entries - 1);
      coeffs = sparsevector<numeric>(fields);
      for (unsigned long i = 1; i <= p.termcount; i++) {
        if (p.terms[i].deg() != degree)
          Error(PolyNotHomo, "homopoly=poly");
        coeffs.append(indets.GetHomoPos(p.terms[i].mono), p.terms[i].coeff);
      }
      leadmono = p.LM();
      return *this;
    }

    homopoly operator=(homopoly<numeric> p) {
      degree   = p.degree  ;
      fields   = p.fields  ;
      coeffs   = p.coeffs  ;
      leadmono = p.leadmono;
      return *this;
    }

    // type conversion:

    operator poly<numeric>() {
      poly<numeric> p;
      sparseentry<numeric>* coeff = coeffs.first;
      while (coeff) {
        p.termcount++;
        p.terms.append(term<numeric>(coeff->value, indets.GetHomoMono(degree, coeff->position)));
        coeff = coeff->next;
      }
      if (p.termcount)
        p.leader = &p.terms[1];
      return p;
    }

    operator poly<numeric>() const {
      poly<numeric> p;
      sparseentry<numeric> *coeff = coeffs.first;
      while (coeff) {
        p.termcount++;
        p.terms.append(term<numeric>(coeff->value, indets.GetHomoMono(degree, coeff->position)));
        coeff = coeff->next;
      }
      if (p.termcount)
        p.leader = &p.terms[1];
      return p;
    }

    // methods:

    // vector
    sparsevector<numeric> coeffvector() const {
      return coeffs;
    }

    // degree
    Z deg() const {
      return degree;
    }

    // leading coefficient
    numeric LC() {
      return leadmono.coeff;
    }

    // leading monomial (which is a "term" here)
    term<numeric> LM() {
      return leadmono;
    }

    // leading term (which is a "monomial" here)
    monomial LT() {
      return leadmono.mono;
    }

    // normal reduction
    homopoly NR(const homopoly &p) const {
      return NR(list<homopoly>(1, p));
    }

    homopoly StatNR(const homopoly &p, unsigned long &count) const {
      return StatNR(list<homopoly>(1, p), count);
    }

    homopoly NR(list<homopoly> g) const {
      unsigned long i, s = g.length();
      homopoly v = *this, q, diff, zeropoly;
      term<numeric> termdiv;
      monomial LTv;

      do {
        do {
          LTv = v.LT();
          for (i = 1; i <= s; i++) {
            if (g[i].LT().devides(LTv)) {
              termdiv = v.LM() / g[i].LM();
              diff = termdiv * g[i];
              v -= diff;
              break;
            }
          }
        } while (i <= s && v != zeropoly);
        q += v.LM();
        v -= v.LM();
      } while (v != zeropoly);
      return q;
    }

    homopoly StatNR(list<homopoly> g, unsigned long &count) const {
      unsigned long i, s = g.length();
      homopoly v = *this, q, zeropoly;
      term<numeric> termdiv;
      monomial LTv;

      do {
        do {
          LTv = v.LT();
          for (i = 1; i <= s; i++) {
            if (g[i].LT().devides(LTv)) {
              count++;
              termdiv = v.LM() / g[i].LM();
              v -= termdiv * g[i];
              break;
            }
          }
        } while (i <= s && v != zeropoly);
        q += v.LM();
        v -= v.LM();
      } while (v != zeropoly);
      return q;
    }

    // support
    list<monomial> supp() const {
      list<monomial> support;
      for (unsigned long i = 1; i <= fields; i++)
        if (coeffs[i] != 0)
          support.append(indets.GetHomoMono(degree, i));
      return support;
    }

  private:
    unsigned long         fields  ;
    Z                     degree  ;
    sparsevector<numeric> coeffs  ;
    term        <numeric> leadmono;

    // method:

    // division algorithm
    list<homopoly> Div(const homopoly &p) const {
      return Div(list<homopoly>(1, p));
    }

    list<homopoly> Div(list<homopoly> g) const {
      unsigned long i, q = g.length() + 1;
      poly<numeric> v = *this, zeropoly;
      term<numeric> termdiv;
      monomial LTv;
      list<homopoly> p(q, zeropoly);

      do {
        do {
          LTv = v.LT();
          for (i = 1; i < q; i++) {
            if (g[i].LT().devides(LTv)) {
              termdiv = v.LM() / g[i].LM();
              p[i] += termdiv       ;
              v    -= termdiv * g[i];
              break;
            }
          }
        } while (i < q);
        p[q] += v.LM();
        v    -= v.LM();
      } while (v != zeropoly);
      return p;
    }

    // Leading Term:

    void findLM() {
      if (coeffs.first)
        leadmono = term<numeric>(coeffs.first->value, indets.GetHomoMono(degree, coeffs.first->position));
      else
        leadmono = term<numeric>(0);
    }
};

// external operators:

// addition and substraction operators:

template<class numeric> homopoly<numeric> operator+(const homopoly<numeric> &p) {
  return p;
}

template<class numeric> homopoly<numeric> operator-(const homopoly<numeric> &p) {
  homopoly<numeric> temp = p;
  temp.coeffs   = -temp.coeffs  ;
  temp.leadmono = -temp.leadmono;
  return temp;
}

template<class numeric> poly<numeric> operator+(const homopoly<numeric> &p, const numeric &a) {
  poly<numeric> temp;
  temp = p;
  return temp += a;
}

template<class numeric> poly<numeric> operator+(const numeric &a, const homopoly<numeric> &p) {
  poly<numeric> temp = a;
  return temp += p;
}

template<class numeric> poly<numeric> operator+(const homopoly<numeric> &p, const pp &x) {
  poly<numeric> temp;
  temp = p;
  return temp += x;
}

template<class numeric> poly<numeric> operator+(const pp &x, const homopoly<numeric> &p) {
  poly<numeric> temp = x;
  return temp += p;
}

template<class numeric> poly<numeric> operator+(const homopoly<numeric> &p, const monomial &m) {
  poly<numeric> temp;
  temp = p;
  return temp += m;
}

template<class numeric> poly<numeric> operator+(const monomial &m, const homopoly<numeric> &p) {
  poly<numeric> temp = m;
  return temp += p;
}

template<class numeric> poly<numeric> operator+(const homopoly<numeric> &p, const term<numeric> &t) {
  poly<numeric> temp;
  temp = p;
  return temp += t;
}

template<class numeric> poly<numeric> operator+(const term<numeric> &t, const homopoly<numeric> &p) {
  poly<numeric> temp = t;
  return temp += p;
}

template<class numeric> poly<numeric> operator+(const homopoly<numeric> &p, const homopoly<numeric> &q) {
  poly<numeric> temp;
  temp = p;
  return temp += q;
}

template<class numeric> poly<numeric> operator+(const poly<numeric> &p, const homopoly<numeric> &h) {
  poly<numeric> temp = p;
  return temp += h;
}

template<class numeric> poly<numeric> operator+(const homopoly<numeric> &h, const poly<numeric> &p) {
  poly<numeric> temp;
  temp = h;
  return temp += p;
}

template<class numeric> poly<numeric> operator-(const homopoly<numeric> &p, const numeric &a) {
  poly<numeric> temp;
  temp = p;
  return temp -= a;
}

template<class numeric> poly<numeric> operator-(const numeric &a, const homopoly<numeric> &p) {
  poly<numeric> temp = a;
  return temp -= p;
}

template<class numeric> poly<numeric> operator-(const homopoly<numeric> &p, const pp &x) {
  poly<numeric> temp;
  temp = p;
  return temp -= x;
}

template<class numeric> poly<numeric> operator-(const pp &x, const homopoly<numeric> &p) {
  poly<numeric> temp = x;
  return temp -= p;
}

template<class numeric> poly<numeric> operator-(const homopoly<numeric> &p, const monomial &m) {
  poly<numeric> temp;
  temp = p;
  return temp -= m;
}

template<class numeric> poly<numeric> operator-(const monomial &m, const homopoly<numeric> &p) {
  poly<numeric> temp = m;
  return temp -= p;
}

template<class numeric> poly<numeric> operator-(const homopoly<numeric> &p, const term<numeric> &t) {
  poly<numeric> temp;
  temp = p;
  return temp -= t;
}

template<class numeric> poly<numeric> operator-(const term<numeric> &t, const homopoly<numeric> &p) {
  poly<numeric> temp = t;
  return temp -= p;
}

template<class numeric> poly<numeric> operator-(const homopoly<numeric> &p, const homopoly<numeric> &q) {
  poly<numeric> temp;
  temp = p;
  return temp -= q;
}

template<class numeric> poly<numeric> operator-(const poly<numeric> &p, const homopoly<numeric> &h) {
  poly<numeric> temp = p;
  return temp -= h;
}

template<class numeric> poly<numeric> operator-(const homopoly<numeric> &h, const poly<numeric> &p) {
  poly<numeric> temp;
  temp = h;
  return temp -= p;
}

// multiplication and division operators:

template<class numeric> homopoly<numeric> operator*(const homopoly<numeric> &p, const numeric &a) {
  homopoly<numeric> temp = p;
  return temp *= a;
}

template<class numeric> homopoly<numeric> operator*(const numeric &a, const homopoly<numeric> &p) {
  homopoly<numeric> temp = p;
  return temp *= a;
}

template<class numeric> homopoly<numeric> operator*(const homopoly<numeric> &p, const monomial &m) {
  homopoly<numeric> temp = p;
  return temp *= m;
//  term    <numeric> q    = term<numeric>(1, m);
//  return temp *= q;
}

template<class numeric> homopoly<numeric> operator*(const monomial &m, const homopoly<numeric> &p) {
  homopoly<numeric> temp = p;
  return temp *= m;
//  term    <numeric> q    = term<numeric>(1, m);
//  return temp *= q;
}

template<class numeric> homopoly<numeric> operator*(const homopoly<numeric> &p, const term<numeric> &t) {
  homopoly<numeric> temp = p;
  homopoly<numeric> q    = t;
  return temp *= q;
}

template<class numeric> homopoly<numeric> operator*(const term<numeric> &t, const homopoly<numeric> &p) {
  homopoly<numeric> temp = p;
  homopoly<numeric> q    = t;
  return temp *= q;
}

template<class numeric> homopoly<numeric> operator*(const homopoly<numeric> &p, const homopoly<numeric> &q) {
  homopoly<numeric> temp = p;
  return temp *= q;
}

template<class numeric> poly<numeric> operator*(const poly<numeric> &p, const homopoly<numeric> &h) {
  poly<numeric> temp = p;
  return temp *= h;
}

template<class numeric> poly<numeric> operator*(const homopoly<numeric> &h, const poly<numeric> &p) {
  poly<numeric> temp = p;
  return temp *= h;
}

template<class numeric> homopoly<numeric> operator/(const homopoly<numeric> &p, const numeric &a) {
  homopoly<numeric> temp = p;
  return temp /= a;
}

template<class numeric> homopoly<numeric> operator/(const homopoly<numeric> &p, const pp &x) {
  homopoly<numeric> temp = p;
  return temp /= x;
}

template<class numeric> homopoly<numeric> operator/(const homopoly<numeric> &p, const monomial &m) {
  homopoly<numeric> temp = p;
  return temp /= m;
}

template<class numeric> homopoly<numeric> operator/(const homopoly<numeric> &p, const term<numeric> &t) {
  homopoly<numeric> temp = p;
  return temp /= t;
}

template<class numeric> homopoly<numeric> operator/(const homopoly<numeric> &p, const homopoly<numeric> &q) {
  homopoly<numeric> temp = p;
  return temp /= q;
}

template<class numeric> homopoly<numeric> operator/(const homopoly<numeric> &h, const poly<numeric> &p) {
  homopoly<numeric> temp = h;
  return temp /= p;
}


template<class numeric> homopoly<numeric> operator%(const homopoly<numeric> &p, const numeric &a) {
  homopoly<numeric> temp = p;
  return temp %= a;
}

template<class numeric> homopoly<numeric> operator%(const homopoly<numeric> &p, const pp &x) {
  homopoly<numeric> temp = p;
  return temp %= x;
}

template<class numeric> homopoly<numeric> operator%(const homopoly<numeric> &p, const monomial &m) {
  homopoly<numeric> temp = p;
  return temp %= m;
}

template<class numeric> homopoly<numeric> operator%(const homopoly<numeric> &p, const term<numeric> &t) {
  homopoly<numeric> temp = p;
  return temp %= t;
}

template<class numeric> homopoly<numeric> operator%(const homopoly<numeric> &p, const homopoly<numeric> &q) {
  homopoly<numeric> temp = p;
  return temp %= q;
}

template<class numeric> homopoly<numeric> operator^(const homopoly<numeric> &p, const Z &a) {
  poly<numeric> temp(p);
  temp ^= a;
  return homopoly<numeric>(temp);
}

// comparison operators:

template<class numeric> bool operator==(const homopoly<numeric> &p, const homopoly<numeric> &q) {
  return p.degree == q.degree && p.coeffs == q.coeffs;
}

template<class numeric> bool operator!=(const homopoly<numeric> &p, const homopoly<numeric> &q) {
  return !(p == q);
}

template<class numeric> bool operator==(const poly<numeric> &p, const homopoly<numeric> &h) {
  return p == poly<numeric>(h);
}

template<class numeric> bool operator!=(const poly<numeric> &p, const homopoly<numeric> &h) {
  return !(p == h);
}

template<class numeric> bool operator==(const homopoly<numeric> &h, const poly<numeric> &p) {
  return p == poly<numeric>(h);
}

template<class numeric> bool operator!=(const homopoly<numeric> &h, const poly<numeric> &p) {
  return !(p == h);
}

template<class numeric> bool operator<(homopoly<numeric> &p, homopoly<numeric> &q) {
  if (p == homopoly<numeric>(0) && q == homopoly<numeric>(0))
    return false;
  if (p.LM() == q.LM()) {
    homopoly<numeric> pp(p), qq(q);
    pp -= pp.LM();
    qq -= qq.LM();
    return pp < qq;
  }
  return p.LM() < q.LM();
}

template<class numeric> bool operator>(homopoly<numeric> &p, homopoly<numeric> &q) {
  return q < p;
}

template<class numeric> bool operator<=(homopoly<numeric> &p, homopoly<numeric> &q) {
  return p == q || p < q;
}

template<class numeric> bool operator>=(homopoly<numeric> &p, homopoly<numeric> &q) {
  return p == q || q < p;
}

// output operator:

template<class numeric> ostream& operator<<(ostream &out, const homopoly<numeric> &p) {
  sparseentry<numeric> *entry      = p.coeffs.first;
  bool                 termprinted = false         ;
  if (entry) {
    while (entry) {
      if (entry->value < 0) {
        if (termprinted)
          out << ' ';
        out << "- ";
        if (entry->value != -1)
          out << -entry->value;
      }

      else {
        if (termprinted)
          out << " + ";
        if (entry->value != 1)
          out << entry->value;
      }
      out << indets.GetHomoMono(p.degree, entry->position);
      termprinted = true      ;
      entry       = entry->next;
    }
  }
  else
    out << '0';
  return out;
}

// functions:

template<class numeric> homopoly<numeric> normal(homopoly<numeric> &p) {
  numeric leadcoeff = p.LC();
  homopoly<numeric> temp = p / leadcoeff;
  if (leadcoeff * temp == p)
    return temp;
  return p;
}


/**********/
/* fields */
/**********/

// list of all available fields for

enum FIELD {
  none,
  RAT ,
  MOD
};

// active field:

extern FIELD Field;


/***************/
/* term orders */
/***************/

// list of special term orders

enum TO {
  Ord      , // not implemented yet
  Lex      ,
  DegLex   ,
  DegRevLex,
  Elim
};

// active term order:

extern TO            TermOrd    ;
extern unsigned long TermOrdElim;

// function for setting active term order:

void termorder(const TO&, const unsigned long &n = 0);

#endif

